/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file haralick.cc
   \brief 2D & 3D Haralick features.
	
	This file contains functions for calculating 2D/3D Haralick features.

	Haralick, Robert M. Shanmugam, K. Dinstein, Its'Hak; 
	Textural Features for Image Classification; 
	IEEE Transactions on Systems, Man and Cybernetics; 
	Publication Date: Nov. 1973; Volume: 3, Issue: 6; On page(s): 610-621; 
	ISSN: 0018-9472
	
	Tesa, L., Smutek, D., Shimizu, A., and Kobatake, H. 
	3D extension of Haralick texture features for medical image analysis. 
	In Proceedings of the Fourth Conference 
	on IASTED international Conference: Signal Processing, Pattern Recognition, 
	and Applications (Innsbruck, Austria, February 14 - 16, 2007). R. Sablatnig 
	and O. Scherzer, Eds. ACTA Press, Anaheim, CA, 350-355.
	
   \author David Svoboda <svoboda@fi.muni.cz>
	\author Jana Matejkova <256364@mail.muni.cz>
	\date 2010
*/

#include <stdlib.h>
#include <iomanip>

#include "transform.h"
#include "haralick.h"

#define SQR(x) ((x)*(x))

using namespace std;

#ifdef WITH_LAPACK
	/// LAPACK function for calculating eigenvalues and eigenvectors
	extern "C" void dgeev_(const char *jobvl, const char *jobvr, int *n, double *a,
						   int *lda, double *wr, double *wi, double *vl,
						   int *ldvl, double *vr, int *ldvr,
						   double *work, int *lwork, int *info);
#endif

namespace i3d {

//----------------------------------------------------------------------------
// Verify the user input
//----------------------------------------------------------------------------
template <class T> 
void HaralickFeatures<T>::TestInputs()
{
	 // test the input parameters for validity
	 if (quantization_level > std::numeric_limits<T>::max())
	 {
		  throw "HaralickFeatures::TestInputs() "
					 "... a strange value of quantization constant.";
	 }

	 if ((typeid(T) == typeid(float)) ||
		  (typeid(T) == typeid(double)))
	 {
		  throw "HaralickFeatures::TestInputs() "
					 " cannot manage floating point data.";
	 }

}

//----------------------------------------------------------------------------
// Prepare the image for further analysis
//----------------------------------------------------------------------------
template <class T>
void HaralickFeatures<T>::ImagePreprocessing(const i3d::VOI<i3d::PIXELS> &voi)
{
	 // First, we must resample the input image 
	 // to guarantee the isotropic resolution.
	 // This operation must be performed as the first because the resampling
	 // may change the range of intensities. The min a max intensity levels
	 // are important for further computation.
	 Vector3d<float> res = analyzed_image.GetResolution().GetRes();

	 // 2D-image
	 if (analyzed_image.GetSizeZ() == 1)
	 {
		  if (res.x != res.y)
		  {
#ifdef I3D_DEBUG
			  cout << "2D resampling required." << endl;
#endif
			  float max_res = std::max(res.x, res.y);
			  Vector3d<float> new_res(max_res, max_res, res.z);

			  ResampleToDesiredResolution(analyzed_image, new_res, LANCZOS);
		  }
	 }
	 else // 3D-image
	 {
		  if ((res.x != res.y) || (res.x != res.z))
		  {
#ifdef I3D_DEBUG
		  		cout << "3D resampling required." << endl;
#endif
		  		float max_res = std::max(std::max(res.x, res.y), res.z);
		  		Vector3d<float> new_res = max_res;

		  		ResampleToDesiredResolution(analyzed_image, new_res, LANCZOS);
		  }
	 }

	 double modifier;

	 // Do we want to use full bit depth?
	 if (use_full_bitdepth)
	 {
		  if (quantization_level == 0)
				quantization_level = std::numeric_limits<T>::max() + 1;

		  const double num_of_colors = std::numeric_limits<T>::max() + 1.0;
		  modifier = ((double) quantization_level) / num_of_colors;
	 }
	 else
	 {
		  if (quantization_level == 0)
				quantization_level = analyzed_image.GetVoxelData().max() + 1;

		  const double num_of_colors_used = analyzed_image.GetVoxelData().max() + 1.0;
		  modifier = ((double) quantization_level) / num_of_colors_used;
	 }

#ifdef I3D_DEBUG
	 cout << "quantization level: " << quantization_level << endl;
	 cout << "modifier: " << modifier << endl;
	 cout << "max image intensity: " << 
				(size_t) analyzed_image.GetVoxelData().max() << endl;
#endif

	 // modify the input image data (stretch and quantize)

	 for (size_t j=0; j<analyzed_image.GetImageSize(); j++)
	 {
		  T value = (T) floor(static_cast<double>(analyzed_image.GetVoxel(j)) * modifier);
		  analyzed_image.SetVoxel(j, value);
	 }

	 // crop if required
	 if (!voi.is_empty())
	 {
		  Image3d<T> subimage(analyzed_image, &voi);
		  analyzed_image = subimage;
	 }

	/* 
#ifdef I3D_DEBUG
	 cout << "Quantized image:" << endl;
	 for (size_t z=0; z<analyzed_image.GetSizeZ(); z++)
	 {
		  for (size_t y=0; y<analyzed_image.GetSizeY(); y++)
		  {
				for (size_t x=0; x<analyzed_image.GetSizeX(); x++)
				{
					 cout << "[" << setw(3) 
								<< (size_t) analyzed_image.GetVoxel(x,y,z) << "]";
				}
				cout << endl;
		  }
		  cout << endl;
	 }
#endif
*/


}

//----------------------------------------------------------------------------
// memory management
//----------------------------------------------------------------------------

template <class T>
void HaralickFeatures<T>::MemoryAllocation()
{
	 cooc_matrix = new std::valarray<matrix_row>[amount];
	 mi_x = new double[amount];
	 mi_y = new double[amount];
	 var_x = new double[amount];
	 var_y = new double[amount];
	 HXY = new double[amount];
	 HX = new double[amount];
	 HY = new double[amount];
	 HXY1 = new double[amount];
	 HXY2 = new double[amount];
}

//----------------------------------------------------------------------------
// math computing
//----------------------------------------------------------------------------
template <class T>
void HaralickFeatures<T>::MathPreprocessing()
{
	 // Apply the preprocessing step for each separation vector individually
	 for (size_t order = 0; order < amount; order++)
	 { 
		  /// create cooccurrence matrices
		  CreateCoocMatrix(order, analyzed_image);

		  /// calculating statistics
		  PrecalculateStatistics(order);
	 }
}

//----------------------------------------------------------------------------
// constructor
//----------------------------------------------------------------------------

template <class T> HaralickFeatures<T>::HaralickFeatures(
		const i3d::Image3d<T> &image, 
		const i3d::Vector3d<int> *_separation_vectors, 
		size_t _amount,
		size_t _quantization_level,
		const i3d::VOI<i3d::PIXELS> &voi,
		bool _use_full_bitdepth) :
		  separation_vectors(_separation_vectors),
		  amount(_amount),
		  quantization_level(_quantization_level),
		  use_full_bitdepth(_use_full_bitdepth)
{
#ifdef I3D_DEBUG
	 cout << "amount of obtained separation vectors: " << amount << endl;
#endif

	 analyzed_image = image;

	 TestInputs();
	 MemoryAllocation();
	 ImagePreprocessing(voi);
	 MathPreprocessing();
}

//----------------------------------------------------------------------------
// constructor
//----------------------------------------------------------------------------

template <class T> HaralickFeatures<T>::HaralickFeatures(
		const i3d::Image3d<T> &image, 
		const i3d::Vector3d<int> &_separation_vector, 
		size_t _quantization_level,
		const i3d::VOI<i3d::PIXELS> &voi,
		bool _use_full_bitdepth):
		  separation_vectors(&_separation_vector),
		  amount(1),
		  quantization_level(_quantization_level),
		  use_full_bitdepth(_use_full_bitdepth)

{
	 analyzed_image = image;

	 TestInputs();
	 MemoryAllocation();
	 ImagePreprocessing(voi);
	 MathPreprocessing();
}
	
//----------------------------------------------------------------------------
// method responsible for creation of cooccurrence matrix
//----------------------------------------------------------------------------
template <class T> 
void HaralickFeatures<T>::CreateCoocMatrix(size_t order,
														 const i3d::Image3d<T> &image)
{
	 // the separation vector must be smaller than the processed (sub)image
	 if (!le(abs(separation_vectors[order]), (Vector3d<int>) image.GetSize()))
	 {
		  throw "HaralickFeatures::CreateCoocMatrix() "
					 "... Bad separation vector size!";
	 }

	 // fill the cooccurrence matrix with the appropriate values
	 const Vector3d<size_t> low = 
				- min(Vector3d<int>(0), separation_vectors[order]);

	 const Vector3d<size_t> high = 
				image.GetSize() - 
				(Vector3d<size_t>)max(Vector3d<int>(0), separation_vectors[order]);

	 size_t count = 0;

    // allocate the cooccurrence matrix and fill it with zeros
	 cooc_matrix[order].resize(quantization_level); 
	 
	 for (size_t i=0; i<quantization_level; i++)
	 {
		  cooc_matrix[order][i].resize(quantization_level, 0); 
	 } 

	 for (size_t z = low.z; z < high.z; z++)
		  for (size_t y = low.y; y < high.y; y++)
				for (size_t x = low.x; x < high.x; x++)
				{
					 Vector3d<size_t> 
								curr_pos(x,y,z),
								next_pos = curr_pos + separation_vectors[order];

					 T curr_val = image.GetVoxel(curr_pos),
						next_val = image.GetVoxel(next_pos);

					 if (curr_val >= quantization_level)
						 cout << "#### " << curr_val << endl;

					 if (next_val >= quantization_level)
						 cout << "#### " << next_val << endl;

					 cooc_matrix[order][curr_val][next_val] += 1.0;
					 // cooc_matrix[next_val][curr_val] += 1.0;

					 count += 1;
				}

	 for (size_t i=0; i<quantization_level; i++)
		  for (size_t j=0; j<quantization_level; j++)
		  {
				cooc_matrix[order][i][j] /= count;
		  }

#ifdef I3D_DEBUG
	 double sum = 0.0;
	 for (size_t i=0; i<quantization_level; i++)
		  for (size_t j=0; j<quantization_level; j++)
				sum += cooc_matrix[order][i][j];

	 cout << "Sum of cooc_matrix[" << order << "] is: " << sum << endl;
#endif
}
	
//----------------------------------------------------------------------------
/// calculate statistics
//----------------------------------------------------------------------------
template <class T> 
void HaralickFeatures<T>::PrecalculateStatistics(size_t order)
{
		// mi_x
		mi_x[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) {
			for (size_t j=1; j<=quantization_level; j++) {
				mi_x[order] += (i-1)*cooc_matrix[order][i-1][j-1];
			}
		}

		// mi_y
		mi_y[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) {
			for (size_t j=1; j<=quantization_level; j++) {
				mi_y[order] += (j-1)*cooc_matrix[order][i-1][j-1];
			}
		}

		// var_x
		var_x[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) {
			for (size_t j=1; j<=quantization_level; j++) {
				var_x[order] += 
						  (i-1-mi_x[order])*
						  (i-1-mi_x[order])*
						  cooc_matrix[order][i-1][j-1];
			}
		}

		// var_y
		var_y[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) {
			for (size_t j=1; j<=quantization_level; j++) {
				var_y[order] += 
						  (j-1-mi_y[order])*
						  (j-1-mi_y[order])*
						  cooc_matrix[order][i-1][j-1];
			}
		}

		// HXY
		HXY[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				if (cooc_matrix[order][i-1][j-1]>0) 
				{
					HXY[order] += 
							  cooc_matrix[order][i-1][j-1] *
							  log10(cooc_matrix[order][i-1][j-1]);
				} 
			}
		HXY[order] *= -1;

		// HX
		HX[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) 
		{
			if (c_x(order, i) > 0) 
			{
				HX[order] += c_x(order, i) * log10(c_x(order, i));
			} 
		}
		HX[order] *= -1;

		// HY
		HY[order] = 0.0;
		for (size_t j=1; j<=quantization_level; j++) 
		{
			if (c_y(order, j) > 0) 
			{ 
				HY[order] += c_y(order, j) * log10(c_y(order, j));
			} 
		}
		HY[order] *= -1;
		
		// HXY1
		HXY1[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				if (c_x(order, i)*c_y(order, j) > 0) 
				{ 
					HXY1[order] += 
							  cooc_matrix[order][i-1][j-1]*
							  log10(c_x(order, i)*c_y(order, j));
				} 
			}
		HXY1[order] *= -1;

		// HXY2
		HXY2[order] = 0.0;
		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				if (c_x(order, i)*c_y(order, j) > 0) 
				{
					HXY2[order] += 
							  c_x(order, i)*
							  c_y(order, j)*
							  log10(c_x(order, i)*c_y(order, j));
				} 
			}
		HXY2[order] *= -1;
}
	
//----------------------------------------------------------------------------
/// function for calculating line sum \f$c_x(i)\f$
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::c_x(size_t order, size_t i) const
{
		double cx = 0.0;
		for (size_t j=1; j<=quantization_level; j++) {
			cx += cooc_matrix[order][i-1][j-1];
		}
		return cx;
}

//----------------------------------------------------------------------------
/// function for calculating line sum \f$c_y(j)\f$
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::c_y(size_t order, size_t j) const
{
		double cy = 0.0;
		for (size_t i=1; i<=quantization_level; i++) {
			cy += cooc_matrix[order][i-1][j-1];
		}
		return cy;
}

//----------------------------------------------------------------------------
/// function for calculating \f$c_x_plus_y(k)\f$
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::c_x_plus_y(size_t order, size_t k) const
{
		double c_xplusy = 0.0;
		for (size_t j=1; j<=quantization_level; j++) {
			for (size_t i=1; i<=quantization_level; i++) {
				if (i+j == k-1) 
					c_xplusy += cooc_matrix[order][i-1][j-1];
			}
		}
		return c_xplusy;
}

//----------------------------------------------------------------------------
/// function for calculating \f$c_x_minus_y(k)\f$
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::c_x_minus_y(size_t order, size_t k) const
{
		double c_xminusy = 0.0;
		for (size_t j=1; j<=quantization_level; j++) {
			for (size_t i=1; i<=quantization_level; i++) {
				if ((size_t)std::abs((int)i-(int)j) == k-1) 
					c_xminusy += cooc_matrix[order][i-1][j-1];
			}
		}
		return c_xminusy;
}

/* HARALICK FEATURES */
//----------------------------------------------------------------------------
// Entropy
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetEntropy(size_t order) const
{
		return HXY[order];
}

//----------------------------------------------------------------------------
// TextureContrast
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetTextureContrast(size_t order) const
{
		double texture_contrast = 0.0;

		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				texture_contrast += 
						  std::abs((int)i-(int)j) * cooc_matrix[order][i-1][j-1];
			}

		return texture_contrast;
}

//----------------------------------------------------------------------------
// TextureCorrelation
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetTextureCorrelation(size_t order) const
{
		double texture_correlation = 0.0;

		if ((var_x[order]*var_y[order]) > 0) 
		{
			for (size_t i=1; i<=quantization_level; i++) 
				for (size_t j=1; j<=quantization_level; j++) 
				{
					texture_correlation += 
							  ((i-1-mi_x[order])*
								(j-1-mi_y[order])*
								cooc_matrix[order][i-1][j-1]);
				}

			texture_correlation /= sqrt(var_x[order]*var_y[order]);
		}

		return texture_correlation;
}

//----------------------------------------------------------------------------
// TextureHomogeneity
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetTextureHomogeneity(size_t order) const
{
		double texture_homogeneity = 0.0;

		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				texture_homogeneity += 
						  cooc_matrix[order][i-1][j-1] / (1+std::abs((int)i-(int)j));
			}

		return texture_homogeneity;
}

//----------------------------------------------------------------------------
// InverseDifferenceMoment
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetInverseDifferenceMoment(size_t order) const
{
		double inverse_difference_moment = 0.0;

		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				if (i != j) 
				{ 
					inverse_difference_moment += 
							  cooc_matrix[order][i-1][j-1]/std::abs((int)i-(int)j);
				}
			}

		return inverse_difference_moment;
}

//----------------------------------------------------------------------------
// MaximumProbability
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetMaximumProbability(size_t order) const
{
		double maximum_probability = 0.0;

		for (size_t i=1; i<=quantization_level; i++)
			for (size_t j=1; j<=quantization_level; j++) 
			{
				if (maximum_probability < cooc_matrix[order][i-1][j-1]) 
				{
					maximum_probability = cooc_matrix[order][i-1][j-1];
				}
			}

		return maximum_probability;
}

//----------------------------------------------------------------------------
// UniformityOfEnergy
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetUniformityOfEnergy(size_t order) const
{
		double uniformity_of_energy = 0.0;

		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				uniformity_of_energy += SQR(cooc_matrix[order][i-1][j-1]);
			}

		return uniformity_of_energy;
}

//----------------------------------------------------------------------------
// Contrast
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetContrast(size_t order) const
{
		double contrast = 0.0;

		for (size_t k=1; k<=quantization_level; k++) 
		{
				contrast += SQR(k-1)*c_x_minus_y(order, k);
		}

		return contrast;
}

//----------------------------------------------------------------------------
// SumAverage
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetSumAverage(size_t order) const
{
		double sum_average = 0.0;

		for (size_t k=1; k<=2*quantization_level-1; k++) 
		{
				sum_average += (k-1) * c_x_plus_y(order, k);
		}

		return sum_average;
}

//----------------------------------------------------------------------------
// SumEntropy
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetSumEntropy(size_t order) const
{
		double sum_entropy = 0.0;

		for (size_t k=1; k<=2*quantization_level-1; k++) 
		{
			if (c_x_plus_y(order, k) != 0) 
			{ 
				sum_entropy += c_x_plus_y(order, k) * log10(c_x_plus_y(order, k));
			} 
		}

		sum_entropy *= -1;

		return sum_entropy;
}

//----------------------------------------------------------------------------
// DifferenceEntropy
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetDifferenceEntropy(size_t order) const
{
		double difference_entropy = 0.0;

		for (size_t k=1; k<=quantization_level; k++) 
		{
			if (c_x_minus_y(order, k)!=0) 
			{
				difference_entropy += 
						  c_x_minus_y(order, k) * log10(c_x_minus_y(order, k));
			} 
		}

		difference_entropy *= -1;

		return difference_entropy;
}

//----------------------------------------------------------------------------
// Variance
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetVariance(size_t order) const
{
		return var_x[order];
}
		
//----------------------------------------------------------------------------
// InfMeasureOfCorrelationA
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetInfMeasureOfCorrelationA(size_t order) const
{
		double CorrelationA = 0.0;

		if (std::max(HX[order],HY[order]) != 0) 
		{
			CorrelationA = 
					  (HXY[order] - HXY1[order]) / std::max(HX[order],HY[order]);
		}

		return CorrelationA;
}

//----------------------------------------------------------------------------
// InfMeasureOfCorrelationB
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetInfMeasureOfCorrelationB(size_t order) const
{
		double CorrelationB = 0.0;

		CorrelationB = sqrt(1-exp((-2)*(HXY2[order]-HXY[order])));

		return CorrelationB;
}

//----------------------------------------------------------------------------
// MaximalCorrelationCoefficient
//----------------------------------------------------------------------------
template <class T> 
double HaralickFeatures<T>::GetMaximalCorrelationCoefficient(size_t order) const
{   
	#ifdef WITH_LAPACK

		// allocate - Q matrix (Fortran-like indexed [cols x rows])
		double *Q = new double [quantization_level*quantization_level];

		// fill with zeros
		for (size_t i=0; i<quantization_level*quantization_level; i++) 
		{
			 Q[i]=0.0;
		}

		// calculate Q entries
		for (size_t i=1; i<=quantization_level; i++) 
			for (size_t j=1; j<=quantization_level; j++) 
			{
				 double entry = 0.0;
				 size_t entry_index = ((j-1)*quantization_level) + (i-1); 
				 double denominator = c_x(order, i)*c_y(order, j);

				 if (denominator != 0) 
				 {
					  for (size_t k=1; k<=quantization_level; k++) 
					  {
							entry += cooc_matrix[order][i-1][k-1] * 
									  cooc_matrix[order][k-1][j-1];
					  }

					  entry /= denominator;
				 }

				 Q[entry_index] = entry;
			}

		// calculating the eigenvalues

		int  dim=quantization_level, 
			  lwork=4*quantization_level, 
			  info;
		double *eigenvalr = new double[quantization_level];
		double *eigenvali = new double[quantization_level];
		double *eigenvecl = new double[dim*dim];
		double *eigenvecr = new double[dim*dim];
		double *work	  = new double[lwork];
	
		dgeev_("N", "N", &dim, Q, &dim, eigenvalr, eigenvali,
				 eigenvecl, &dim, eigenvecr, &dim, work, &lwork, &info);
		
		double result_is_the_second_eigenvalue=eigenvalr[1];
		
		// free the memory
		delete [] eigenvalr;
		delete [] eigenvali;
		delete [] eigenvecl;
		delete [] eigenvecr;
		delete [] work;
		delete [] Q;
		
		return result_is_the_second_eigenvalue;

	#else
		
		throw InternalException("GetMaximalCorrelationCoefficient: You have selected i3dlibs library without LAPACK support!");

	#endif // WITH_LAPACK
}
	
//-----------------------------------------------------------------------
// public output methods
//-----------------------------------------------------------------------

template <class T> double 
HaralickFeatures<T>::GetFeature(ListOfHaralickFeatures name,
										  size_t order) const
{
	 switch (name)
	 {
		  case Entropy:
					 return GetEntropy(order);
					 break;
		  case TextureContrast:
					 return GetTextureContrast(order);
					 break;
		  case TextureCorrelation:
					 return GetTextureCorrelation(order);
					 break;
		  case TextureHomogeneity:
					 return GetTextureHomogeneity(order);
					 break;
		  case InverseDifferenceMoment:
					 return GetInverseDifferenceMoment(order);
					 break;
		  case MaximumProbability:
					 return GetMaximumProbability(order);
					 break;
		  case UniformityOfEnergy:
					 return GetUniformityOfEnergy(order);
					 break;
		  case Contrast:
					 return GetContrast(order);
					 break;
		  case SumAverage:
					 return GetSumAverage(order);
					 break;
		  case SumEntropy:
					 return GetSumEntropy(order);
					 break;
		  case DifferenceEntropy:
					 return GetDifferenceEntropy(order);
					 break;
		  case Variance:
					 return GetVariance(order);
					 break;
		  case InfMeasureOfCorrelationA:
					 return GetInfMeasureOfCorrelationA(order);
					 break;
		  case InfMeasureOfCorrelationB:
					 return GetInfMeasureOfCorrelationB(order);
					 break;
		  case MaximalCorrelationCoefficient:
					 return GetMaximalCorrelationCoefficient(order);
					 break;
		  default: throw "HaralickFeatures::GetFeature() "
					  "... Unknow Haralick feature.";
	 }
}

//-----------------------------------------------------------------------
template <class T> double 
HaralickFeatures<T>::GetMeanFeature(ListOfHaralickFeatures name) const
{
	 double sum = 0.0;

	 for (size_t order = 0; order < amount; order++)
	 {
		  sum += GetFeature(name, order);
	 }

	 return (sum/amount);
}

//-----------------------------------------------------------------------
// Debug method - this method only shows the separation vectors
//-----------------------------------------------------------------------
template <class T>
void HaralickFeatures<T>::ShowSeparationVectors() const
{
	 for (size_t order = 0; order < amount; order++)
	 {
		  cout << "VECTOR: " << separation_vectors[order] << endl;
	 }
	 cout << endl;
}

//-----------------------------------------------------------------------
// Debug method - this method only shows the cooccurrence matrices
//-----------------------------------------------------------------------
template <class T>
void HaralickFeatures<T>::ShowMatrices() const
{
	 for (size_t order = 0; order < amount; order++)
	 {
		 cout << "Cooccurrence matrix[" << order << "]: " << endl;
		 for (size_t j=0; j<quantization_level; j++)
		 {
			  for (size_t i=0; i<quantization_level; i++)
			  {
					cout << "[" << setw(10) << cooc_matrix[order][i][j] << "]";
			  }
			  cout << endl;
		 }
		 cout << endl;
	 }
}

//----------------------------------------------------------------------------
// destructor
//----------------------------------------------------------------------------
template <class T>
HaralickFeatures<T>::~HaralickFeatures()
{
	 delete [] cooc_matrix;
	 delete [] mi_x;
	 delete [] mi_y;
	 delete [] var_x;
	 delete [] var_y;
	 delete [] HXY;
	 delete [] HX;
	 delete [] HY;
	 delete [] HXY1;
	 delete [] HXY2;
}

//-----------------------------------------------------------------------
// explicit instantiation
//-----------------------------------------------------------------------

template class I3D_ALGO_EXPORT HaralickFeatures<GRAY8>;
template class I3D_ALGO_EXPORT HaralickFeatures<GRAY16>;

}

//-----------------------------------------------------------------------
// end of file 'haralick.cc'
//-----------------------------------------------------------------------
