/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * lab_nha_curvature.cc
 *
 * Curvature used in the labelled versions of the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "lab_nha_curvature.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                   Class Lab_NHA_Curvature                   //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Lab_NHA_Curvature::Estimate(Type::Size i)
		{
			// compute limits of the circular neighbourhood in the x axis
			Type::Size min_x, max_x;
			m_circular_neigh -> LimitsX(i, min_x, max_x);

			// compute limits of the circular neighbourhood in the y axis
			Type::Size min_y, max_y;
			m_circular_neigh -> LimitsY(i, min_y, max_y);

			// shift to the next row of the circular neighbourhood
			Type::Size shift = m_grid.m_sx - max_x + min_x - 1;

			// the first grid point in the circular neighbourhood
			Type::Size start = min_y * m_grid.m_sx + min_x;

			// implicit function at the first grid point in the circular neighbourhood
			const int *ptr = m_grid.m_implicit_function.ptr(start);

			// label at the first grid point in the circular neighbourhood
			const Type::LabVoxel *lptr = m_labels.ptr(start);

			// label at the i-th grid point
			Type::LabVoxel label = m_labels[i];

			// counter of the internal grid points
			int num = 0;

			// go through the circular neighbourhood and look for internal grid points
			for (; min_y <= max_y; ++min_y)
			{
				for (Type::Size x = min_x; x <= max_x; ++x)
				{
					// check whether a grid point is internal
					if (*ptr == -1 && *lptr == label)
					{
						// increment the counter
						++num;
					}					

					++ptr; ++lptr;
				}

				// move to the next row of the circular neighbourhood
				ptr += shift; lptr += shift;
			}

			// update the curvature at the i-th grid point
			m_grid.m_curvature[i] = num;
		}

		/**********************************************************************/

		void Lab_NHA_Curvature::Update(Type::Size i, int value)
		{
			// compute limits of the circular neighbourhood in the x axis
			Type::Size min_x, max_x;
			m_circular_neigh -> LimitsX(i, min_x, max_x);

			// compute limits of the circular neighbourhood in the y axis
			Type::Size min_y, max_y;
			m_circular_neigh -> LimitsY(i, min_y, max_y);
						
			// shift to the next row of the circular neighbourhood
			Type::Size shift = m_grid.m_sx - max_x + min_x - 1;

			// index of the first grid point in the circular neighbourhood
			Type::Size start = min_y * m_grid.m_sx + min_x;

			// curvature at the first grid point in the circular neighbourhood
			int *ptr = m_grid.m_curvature.ptr(start);

			// implicit function at the first grid point in the circular neighbourhood
			const int *imp_ptr = m_grid.m_implicit_function.ptr(start);

			// label at the first grid point in the circular neighbourhood
			const Type::LabVoxel *lptr = m_labels.ptr(start);
			
			// label at the i-th grid point
			Type::LabVoxel label = m_labels[i];

			// go through the circular neighbourhood and update the curvature at interface grid points
			for (; min_y <= max_y; ++min_y)
			{
				for (Type::Size x = min_x; x <= max_x; ++x)
				{
					// check whether a grid point belongs to the interface
					if (*imp_ptr == 0 && *lptr == label)
					{
						// update the curvature by given value
						*ptr += value;
					}

					++ptr;
					++imp_ptr;
					++lptr;
				}

				// move to the next row of the circular neighbourhood
				ptr += shift;
				imp_ptr += shift;
				lptr += shift;
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                  Class Lab_NHA_Curvature3D                  //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Lab_NHA_Curvature3D::Estimate(Type::Size i)
		{
			// compute limits of the circular neighbourhood in the x axis
			Type::Size min_x, max_x;
			m_circular_neigh -> LimitsX(i, min_x, max_x);

			// compute limits of the circular neighbourhood in the y axis
			Type::Size min_y, max_y;
			m_circular_neigh -> LimitsY(i, min_y, max_y);

			// compute limits of the circular neighbourhood in the z axis
			Type::Size min_z, max_z;
			m_circular_neigh -> LimitsZ(i, min_z, max_z);

			// shift to the next row of the circular neighbourhood
			Type::Size row_shift = m_grid.m_sx - max_x + min_x - 1;

			// shift to the next slice of the circular neighbourhood
			Type::Size slice_shift = m_grid.m_slice - (max_y - min_y + 1) * m_grid.m_sx;

			// the first grid point in the circular neighbourhood
			Type::Size start = min_z * m_grid.m_slice + min_y * m_grid.m_sx + min_x;

			// implicit function at the first grid point in the circular neighbourhood
			const int *ptr = m_grid.m_implicit_function.ptr(start);

			// label at the first grid point in the circular neighbourhood
			const Type::LabVoxel *lptr = m_labels.ptr(start);

			// label at the i-th grid point
			Type::LabVoxel label = m_labels[i];

			// counter of the internal grid points
			int num = 0;

			// go through the circular neighbourhood and look for internal grid points
			for (; min_z <= max_z; ++min_z)
			{
				for (Type::Size y = min_y; y <= max_y; ++y)
				{
					for (Type::Size x = min_x; x <= max_x; ++x)
					{
						// check whether a grid point is internal
						if (*ptr == -1 && *lptr == label)
						{
							// increment the counter
							++num;
						}					

						++ptr; ++lptr;
					}

					// move to the next row of the circular neighbourhood
					ptr += row_shift; lptr += row_shift;
				}

				// move to the next slice of the circular neighbourhood
				ptr += slice_shift; lptr += slice_shift;
			}

			// update the curvature at the i-th grid point
			m_grid.m_curvature[i] = num;
		}

		/**********************************************************************/

		void Lab_NHA_Curvature3D::Update(Type::Size i, int value)
		{
			// compute limits of the circular neighbourhood in the x axis
			Type::Size min_x, max_x;
			m_circular_neigh -> LimitsX(i, min_x, max_x);

			// compute limits of the circular neighbourhood in the y axis
			Type::Size min_y, max_y;
			m_circular_neigh -> LimitsY(i, min_y, max_y);

			// compute limits of the circular neighbourhood in the z axis
			Type::Size min_z, max_z;
			m_circular_neigh -> LimitsZ(i, min_z, max_z);

			// shift to the next row of the circular neighbourhood
			Type::Size row_shift = m_grid.m_sx - max_x + min_x - 1;

			// shift to the next slice of the circular neighbourhood
			Type::Size slice_shift = m_grid.m_slice - (max_y - min_y + 1) * m_grid.m_sx;

			// index of the first grid point in the circular neighbourhood
			Type::Size start = min_z * m_grid.m_slice + min_y * m_grid.m_sx + min_x;

			// curvature at the first grid point in the circular neighbourhood
			int *ptr = m_grid.m_curvature.ptr(start);

			// implicit function at the first grid point in the circular neighbourhood
			const int *imp_ptr = m_grid.m_implicit_function.ptr(start);

			// label at the first grid point in the circular neighbourhood
			const Type::LabVoxel *lptr = m_labels.ptr(start);

			// label at the i-th grid point
			Type::LabVoxel label = m_labels[i];

			// go through the circular neighbourhood and update the curvature at interface grid points
			for (; min_z <= max_z; ++min_z)
			{
				for (Type::Size y = min_y; y <= max_y; ++y)
				{
					for (Type::Size x = min_x; x <= max_x; ++x)
					{
						// check whether a grid point belongs to the interface
						if (*imp_ptr == 0 && *lptr == label)
						{
							// update the curvature by given value
							*ptr += value;
						}

						++ptr;
						++imp_ptr;
						++lptr;
					}

					// move to the next row of the circular neighbourhood
					ptr += row_shift;
					imp_ptr += row_shift;
					lptr += row_shift;
				}

				// move to the next slice of the circular neighbourhood
				ptr += slice_shift;
				imp_ptr += slice_shift;
				lptr += slice_shift;
			}
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

