/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nha_circular_neighbourhood.cc
 *
 * Circular neighbourhood used in the Nilsson-Heyden algorithm for approximating the curvature
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "nha_circular_neighbourhood.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//               Class NHA_CircularNeighbourhood               //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		NHA_CircularNeighbourhood::NHA_CircularNeighbourhood(const GridBase &grid, Type::Size radius) : m_grid(grid), m_radius_x(0), m_radius_y(0), m_radius_z(0)
		{
			// the unit of the radius corresponds to the minimal grid point size
			// if the grid point is anisotropic, the radius must be adapted to the grid resolution in each axis and rounded
			Type::Prec max = Max(grid.m_res_x, grid.m_res_y);

			// check whether the computational grid is 3D
			if (grid.m_sz > 1)
			{
				max = Max(max, grid.m_res_z);
				m_radius_z = Type::Size(radius * m_grid.m_res_z / max + Type::Prec(0.5));
			}
			
			m_radius_x = Type::Size(radius * m_grid.m_res_x / max + Type::Prec(0.5));
			m_radius_y = Type::Size(radius * m_grid.m_res_y / max + Type::Prec(0.5));
		}

		/**********************************************************************/

		void NHA_CircularNeighbourhood::LimitsX(Type::Size i, Type::Size &min, Type::Size &max) const
		{
			// compute position of the i-th grid point in the x axis
			Type::Size x = i % m_grid.m_sx;

			// compute limits of the circular neighbourhood in the x axis
			min = x > m_radius_x ? x - m_radius_x : 1;
			max = Min(x + m_radius_x, m_grid.m_sx - 2);
		}

		/**********************************************************************/

		void NHA_CircularNeighbourhood::LimitsY(Type::Size i, Type::Size &min, Type::Size &max) const
		{
			// compute position of the i-th grid point in the y axis
			Type::Size y = (i / m_grid.m_sx) % m_grid.m_sy;

			// compute limits of the circular neighbourhood in the y axis
			min = y > m_radius_y ? y - m_radius_y : 1;
			max = Min(y + m_radius_y, m_grid.m_sy - 2);
		}

		/**********************************************************************/

		void NHA_CircularNeighbourhood::LimitsZ(Type::Size i, Type::Size &min, Type::Size &max) const
		{
			// compute position of the i-th grid point in the z axis
			Type::Size z = i / m_grid.m_slice;

			// compute limits of the circular neighbourhood in the x axis
			min = z > m_radius_z ? z - m_radius_z : 1;
			max = Min(z + m_radius_z, m_grid.m_sz - 2);
		}
		
		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

