/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nha_simple_point_inspector.cc
 *
 * Simple point inspector used in the topology-preserving version of the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "nha_simple_point_inspector.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                      Local definitions                      //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		/** Look-up table for fast simple point detection in 2D considering the 4-adjacency for the background 
		* and 8-adjacency for the foreground. Each value packs 16 configurations. */
		static const unsigned short lut_4b_8f[16] = { 53214, 52445, 52993, 52224, 52993, 13277, 52993, 13277, 
														  1, 52445,     0, 52224, 52993, 13277, 52993, 13277};

		/**********************************************************************/

		/** Add two edges to given graph. */
		#define ADD_2(graph,a,b,val) graph.Edge(a,val); graph.Edge(b,val)

		/**********************************************************************/

		/** Add four edges to given graph. */
		#define ADD_4(graph, a,b,c,d,val) ADD_2(graph,a,b,val); ADD_2(graph,c,d,val)

		/**********************************************************************/

		/** Add six edges to given graph. */
		#define ADD_6(graph,a,b,c,d,e,f,val) ADD_4(graph,a,b,c,d,val); ADD_2(graph,e,f,val)

		/**********************************************************************/

		/** Add ten edges to given graph. */
		#define ADD_10(graph,a,b,c,d,e,f,g,h,i,j,val) ADD_6(graph,a,b,c,d,e,f,val); ADD_4(graph,g,h,i,j,val)

		/**********************************************************************/

		/** Add sixteen edges to given graph. */
		#define ADD_16(graph,a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,val) ADD_10(graph,a,b,c,d,e,f,g,h,i,j,val); ADD_6(graph,k,l,m,n,o,p,val)

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                Class NHA_SimplePointInspector               //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		bool NHA_SimplePointInspector::Verify(Type::Size i)
		{
			if (m_grid.m_implicit_function[i] == 5)
			{
				// the i-th grid point is outside the domain
				return false;
			}

			// represent the i-th grid point configuration as a binary number 
			int val = ((m_grid.m_implicit_function[i - m_grid.m_sx - 1] < 1) << 8) +
					  ((m_grid.m_implicit_function[i - m_grid.m_sx    ] < 1) << 7) +
					  ((m_grid.m_implicit_function[i - m_grid.m_sx + 1] < 1) << 6) +
				      ((m_grid.m_implicit_function[i               - 1] < 1) << 5) +
				      ((m_grid.m_implicit_function[i                  ] < 1) << 4) +
				      ((m_grid.m_implicit_function[i               + 1] < 1) << 3) +
				      ((m_grid.m_implicit_function[i + m_grid.m_sx - 1] < 1) << 2) +
				      ((m_grid.m_implicit_function[i + m_grid.m_sx    ] < 1) << 1) +
				       (m_grid.m_implicit_function[i + m_grid.m_sx + 1] < 1);

			// swap bits in order to invert background and foreground adjacencies
			//val = 511 - val;

			return (((lut_4b_8f[val >> 5]) & (1 << (val & 15))) != 0);
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//               Class NHA_SimplePointInspector3D              //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		bool NHA_SimplePointInspector3D::Verify(Type::Size i)
		{
			if (m_grid.m_implicit_function[i] == 5)
			{
				// the i-th grid point is outside the domain
				return false;
			}

			// create a 26-neighbourhood of the i-th grid point
			m_neigh[0]  = (m_grid.m_implicit_function[i - m_grid.m_slice - m_grid.m_sx - 1] > 0);
			m_neigh[1]  = (m_grid.m_implicit_function[i - m_grid.m_slice - m_grid.m_sx    ] > 0);
			m_neigh[2]  = (m_grid.m_implicit_function[i - m_grid.m_slice - m_grid.m_sx + 1] > 0);
			m_neigh[3]  = (m_grid.m_implicit_function[i - m_grid.m_slice			   - 1] > 0);
			m_neigh[4]  = (m_grid.m_implicit_function[i - m_grid.m_slice				  ] > 0);
			m_neigh[5]  = (m_grid.m_implicit_function[i - m_grid.m_slice		       + 1] > 0);
			m_neigh[6]  = (m_grid.m_implicit_function[i - m_grid.m_slice + m_grid.m_sx - 1] > 0);
			m_neigh[7]  = (m_grid.m_implicit_function[i - m_grid.m_slice + m_grid.m_sx    ] > 0);
			m_neigh[8]  = (m_grid.m_implicit_function[i - m_grid.m_slice + m_grid.m_sx + 1] > 0);
			m_neigh[9]  = (m_grid.m_implicit_function[i				     - m_grid.m_sx - 1] > 0);
			m_neigh[10] = (m_grid.m_implicit_function[i				     - m_grid.m_sx    ] > 0);
			m_neigh[11] = (m_grid.m_implicit_function[i                  - m_grid.m_sx + 1] > 0);
			m_neigh[12] = (m_grid.m_implicit_function[i							       - 1] > 0);
			m_neigh[13] = (m_grid.m_implicit_function[i								   + 1] > 0);
			m_neigh[14] = (m_grid.m_implicit_function[i                  + m_grid.m_sx - 1] > 0);
			m_neigh[15] = (m_grid.m_implicit_function[i                  + m_grid.m_sx    ] > 0);
			m_neigh[16] = (m_grid.m_implicit_function[i                  + m_grid.m_sx + 1] > 0);
			m_neigh[17] = (m_grid.m_implicit_function[i + m_grid.m_slice - m_grid.m_sx - 1] > 0);
			m_neigh[18] = (m_grid.m_implicit_function[i + m_grid.m_slice - m_grid.m_sx    ] > 0);
			m_neigh[19] = (m_grid.m_implicit_function[i + m_grid.m_slice - m_grid.m_sx + 1] > 0);
			m_neigh[20] = (m_grid.m_implicit_function[i + m_grid.m_slice			   - 1] > 0);
			m_neigh[21] = (m_grid.m_implicit_function[i + m_grid.m_slice			      ] > 0);
			m_neigh[22] = (m_grid.m_implicit_function[i + m_grid.m_slice			   + 1] > 0);
			m_neigh[23] = (m_grid.m_implicit_function[i + m_grid.m_slice + m_grid.m_sx - 1] > 0);
			m_neigh[24] = (m_grid.m_implicit_function[i + m_grid.m_slice + m_grid.m_sx    ] > 0);
			m_neigh[25] = (m_grid.m_implicit_function[i + m_grid.m_slice + m_grid.m_sx + 1] > 0);

			return (VerifyBackground() && VerifyForeground());		
		}

		/**********************************************************************/

		bool NHA_SimplePointInspector3D::VerifyBackground()
		{
			// initialize the graph of background
			m_background_graph.Clear();

			// index of the starting vertex
			int start = -1;

			// number of 6-neighbouring vertices
			int num = 0;

			// sum of indices of 6-neighbouring vertices
			int sum = 0;

			if (m_neigh[1]) { ADD_2(m_background_graph,2,6,0); }
			if (m_neigh[3]) { ADD_2(m_background_graph,2,8,1); }
			if (m_neigh[4]) { sum += 2; ++num; start = 2; ADD_4(m_background_graph,0,1,3,4,2); }
			if (m_neigh[5]) { ADD_2(m_background_graph,2,9,3); }
			if (m_neigh[7]) { ADD_2(m_background_graph,2,11,4); }
			if (m_neigh[9]) { ADD_2(m_background_graph,6,8,5); }
			if (m_neigh[10]) { sum += 6; ++num; start = 6; ADD_4(m_background_graph,0,5,7,13,6); }
			if (m_neigh[11]) { ADD_2(m_background_graph,6,9,7); }
			if (m_neigh[12]) { sum += 8; ++num; start = 8; ADD_4(m_background_graph,1,5,10,14,8); }
			if (m_neigh[13]) { sum += 9; ++num; start = 9; ADD_4(m_background_graph,3,7,12,16,9); }
			if (m_neigh[14]) { ADD_2(m_background_graph,8,11,10); }
			if (m_neigh[15]) { sum += 11; ++num; start = 11; ADD_4(m_background_graph,4,10,12,17,11); }
			if (m_neigh[16]) { ADD_2(m_background_graph,9,11,12); }
			if (m_neigh[18]) { ADD_2(m_background_graph,6,15,13); }
			if (m_neigh[20]) { ADD_2(m_background_graph,8,15,14); }
			if (m_neigh[21]) { sum += 15; ++num; start = 15; ADD_4(m_background_graph,13,14,16,17,15); }
			if (m_neigh[22]) { ADD_2(m_background_graph,9,15,16); }
			if (m_neigh[24]) { ADD_2(m_background_graph,11,15,17); }

			switch (num)
			{
				// the situation number one - no 6-neighbouring vertex
			case 0: return false;

				// the situation number two - only one 6-neighbouring vertex
			case 1: return true;

				// the situation number three - two 6-neighbouring vertices
			case 2:

				switch (sum)
				{
					// the first case - two opposite 6-neighbouring vertices
				case 17: return false;

					// the second case = two 6-neighbouring vertices that share one adjacent 18-neighbour
				case  8: return m_neigh[1];
				case 10: return m_neigh[3];
				case 11: return m_neigh[5];
				case 13: return m_neigh[7];
				case 14: return m_neigh[9];
				case 15: return m_neigh[11];
				case 19: return m_neigh[14];
				case 20: return m_neigh[16];
				case 21: return m_neigh[18];
				case 23: return m_neigh[20];
				case 24: return m_neigh[22];
				case 26: return m_neigh[24];
				}

			default: return m_background_graph.AccessibleSixNeighbours(start, num);
			}
		}

		/**********************************************************************/

		bool NHA_SimplePointInspector3D::VerifyForeground()
		{
			// initialize the graph of the foreground
			m_foreground_graph.Clear();

			// index of the starting vertex
			int start = -1;

			// the number of vertices
			int num = 0;

			if (!m_neigh[0]) { ++num; start = 0; ADD_6(m_foreground_graph,1,3,4,9,10,12,0); };
			if (!m_neigh[1]) { ++num; start = 1; ADD_10(m_foreground_graph,0,2,3,4,5,9,10,11,12,13,1); };
			if (!m_neigh[2]) { ++num; start = 2; ADD_6(m_foreground_graph,1,4,5,10,11,13,2); };
			if (!m_neigh[3]) { ++num; start = 3; ADD_10(m_foreground_graph,0,1,4,6,7,9,10,12,14,15,3); };
			if (!m_neigh[4]) { ++num; start = 4; ADD_16(m_foreground_graph,0,1,2,3,5,6,7,8,9,10,11,12,13,14,15,16,4); };
			if (!m_neigh[5]) { ++num; start = 5; ADD_10(m_foreground_graph,1,2,4,7,8,10,11,13,15,16,5); };
			if (!m_neigh[6]) { ++num; start = 6; ADD_6(m_foreground_graph,3,4,7,12,14,15,6); };
			if (!m_neigh[7]) { ++num; start = 7; ADD_10(m_foreground_graph,3,4,5,6,8,12,13,14,15,16,7); };
			if (!m_neigh[8]) { ++num; start = 8; ADD_6(m_foreground_graph,4,5,7,13,15,16,8); };

			if (!m_neigh[9])  { ++num; start =  9; ADD_10(m_foreground_graph,0,1,3,4,10,12,17,18,20,21,9); };
			if (!m_neigh[10]) { ++num; start = 10; ADD_16(m_foreground_graph,0,1,2,3,4,5,9,11,12,13,17,18,19,20,21,22,10); };
			if (!m_neigh[11]) { ++num; start = 11; ADD_10(m_foreground_graph,1,2,4,5,10,13,18,19,21,22,11); };
			if (!m_neigh[12]) { ++num; start = 12; ADD_16(m_foreground_graph,0,1,3,4,6,7,9,10,14,15,17,18,20,21,23,24,12); };
			if (!m_neigh[13]) { ++num; start = 13; ADD_16(m_foreground_graph,1,2,4,5,7,8,10,11,15,16,18,19,21,22,24,25,13); };
			if (!m_neigh[14]) { ++num; start = 14; ADD_10(m_foreground_graph,3,4,6,7,12,15,20,21,23,24,14); };
			if (!m_neigh[15]) { ++num; start = 15; ADD_16(m_foreground_graph,3,4,5,6,7,8,12,13,14,16,20,21,22,23,24,25,15); };
			if (!m_neigh[16]) { ++num; start = 16; ADD_10(m_foreground_graph,4,5,7,8,13,15,21,22,24,25,16); };

			if (!m_neigh[17]) { ++num; start = 17; ADD_6(m_foreground_graph,9,10,12,18,20,21,17); };
			if (!m_neigh[18]) { ++num; start = 18; ADD_10(m_foreground_graph,9,10,11,12,13,17,19,20,21,22,18); };
			if (!m_neigh[19]) { ++num; start = 19; ADD_6(m_foreground_graph,10,11,13,18,21,22,19); };
			if (!m_neigh[20]) { ++num; start = 20; ADD_10(m_foreground_graph,9,10,12,14,15,17,18,21,23,24,20); };
			if (!m_neigh[21]) { ++num; start = 21; ADD_16(m_foreground_graph,9,10,11,12,13,14,15,16,17,18,19,20,22,23,24,25,21); };
			if (!m_neigh[22]) { ++num; start = 22; ADD_10(m_foreground_graph,10,11,13,15,16,18,19,21,24,25,22); };
			if (!m_neigh[23]) { ++num; start = 23; ADD_6(m_foreground_graph,12,14,15,20,21,24,23); };
			if (!m_neigh[24]) { ++num; start = 24; ADD_10(m_foreground_graph,12,13,14,15,16,20,21,22,23,25,24); };
			if (!m_neigh[25]) { ++num; start = 25; ADD_6(m_foreground_graph,13,15,16,21,22,24,25); };

			return (num != 0 && m_foreground_graph.AccessibleNeighbours(start, num));
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

