/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * reg_nha_initializer.cc
 *
 * Initializer of the region-based version of the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "reg_nha_initializer.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                 Class Reg_NHA_Initializer                   //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Reg_NHA_Initializer::Initialize(std::vector<Type::Size> &initial_interface)
		{
			// index of the starting grid point 
			Type::Size index = m_grid.m_sx + 1;

			// implicit function at the starting grid point
			int *ptr = m_grid.m_implicit_function.ptr(index);

			// the first grid point in given mask
			const Type::BinVoxel *mask = m_mask;

			// mask size in the x axis
			Type::Size sx = m_grid.m_sx - 2;

			// mask size in the y axis
			Type::Size sy = m_grid.m_sy - 2;

			// go through the grid and look for border grid points in given mask

			// check the upper left corner
			if (*mask)
			{
				// check whether right and lower neighbours are both in given mask
				if (*(mask + 1) && *(mask + sx))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next grid point
			++index; ++ptr; ++mask;

			// check the remaining grid points in the first row except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				if (*mask)
				{
					// check whether left, right, and lower neighbours are in given mask
					if (*(mask - 1) && *(mask + 1) && *(mask + sx))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;
			}

			// check the upper right corner
			if (*mask)
			{
				// check whether left and lower neighbours are both in given mask
				if (*(mask - 1) && *(mask + sx))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next row
			index += 3; ptr += 3; ++mask;

			// check the remaining rows except the last one
			for (Type::Size y = 2; y < sy; ++y)
			{
				// check the first grid point in the row
				if (*mask)
				{
					// check whether upper, right, and lower neighbours are in given mask
					if (*(mask - sx) && *(mask + 1) && *(mask + sx))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;

				// chech the remaining grid points in the row except the last one
				for (Type::Size x = 2; x < sx; ++x)
				{
					if (*mask)
					{
						// check whether all four neighbours are in given mask
						if (*(mask - sx) && *(mask - 1) && *(mask + 1) && *(mask + sx))
						{
							*ptr = -1;
						}
						else
						{
							*ptr = 0;
							initial_interface.push_back(index);
						}

						// add a grid point to the foreground
						m_obj_manager.AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager.AddBackground(index);
					}

					// move to the next grid point
					++index; ++ptr; ++mask;
				}

				// check the last grid point in the row
				if (*mask)
				{
					// check whether upper, left, and lower neighbours are in given mask
					if (*(mask - sx) && *(mask - 1) && *(mask + sx))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next row
				index += 3; ptr += 3; ++mask;
			}

			// check the lower left corner
			if (*mask)
			{
				// check whether upper and right neighbours are both in given mask
				if (*(mask - sx) && *(mask + 1))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next grid point
			++index; ++ptr; ++mask;

			// check the remaining grid points in the last row except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				if (*mask)
				{
					// check whether upper, left, and right neighbours are in given mask
					if (*(mask - sx) && *(mask - 1) && *(mask + 1))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;
			}

			// check the lower right corner
			if (*mask)
			{
				// check whether upper and left neighbours are both in given mask
				if (*(mask - sx) && *(mask - 1))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                Class Reg_NHA_Initializer3D                  //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		void Reg_NHA_Initializer3D::Initialize(std::vector<Type::Size> &initial_interface)
		{
			// index of the starting grid point 
			Type::Size index = m_grid.m_slice + m_grid.m_sx + 1;

			// implicit function at the starting grid point
			int *ptr = m_grid.m_implicit_function.ptr(index);

			// shift to the next slice
			Type::Size shift = 2 * m_grid.m_sx;

			// the first grid point in given mask
			const Type::BinVoxel *mask = m_mask;

			// mask size in the x axis
			Type::Size sx = m_grid.m_sx - 2;

			// mask size in the y axis
			Type::Size sy = m_grid.m_sy - 2;

			// mask size in the z axis
			Type::Size sz = m_grid.m_sz - 2;

			// mask slice size in the x and y axes
			Type::Size slice = sx * sy;

			// go through the grid and look for border grid points in given mask

			// check the first slice

			// check the upper left corner of the first slice
			if (*mask)
			{
				// check whether right, lower, and bottom neighbours are in given mask
				if (*(mask + 1) && *(mask + sx) && *(mask + slice))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next grid point
			++index; ++ptr; ++mask;

			// check the remaining grid points in the first row of the first slice except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				if (*mask)
				{
					// check whether left, right, lower, and bottom neighbours are in given mask
					if (*(mask - 1) && *(mask + 1) && *(mask + sx) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;
			}

			// check the upper right corner of the first slice
			if (*mask)
			{
				// check whether left, lower, and bottom neighbours are in given mask
				if (*(mask - 1) && *(mask + sx) && *(mask + slice))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next row
			index += 3; ptr += 3; ++mask;

			// check the remaining rows of the first slice except the last one
			for (Type::Size y = 2; y < sy; ++y)
			{
				// check the first grid point in the row
				if (*mask)
				{
					// check whether upper, right, lower, and bottom neighbours are in given mask
					if (*(mask - sx) && *(mask + 1) && *(mask + sx) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;

				// chech the remaining grid points in the row except the last one
				for (Type::Size x = 2; x < sx; ++x)
				{
					if (*mask)
					{
						// check whether upper, left, right, lower, and bottom neighbours are in given mask
						if (*(mask - sx) && *(mask - 1) && *(mask + 1) && *(mask + sx) && *(mask + slice))
						{
							*ptr = -1;
						}
						else
						{
							*ptr = 0;
							initial_interface.push_back(index);
						}

						// add a grid point to the foreground
						m_obj_manager.AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager.AddBackground(index);
					}

					// move to the next grid point
					++index; ++ptr; ++mask;
				}

				// check the last grid point in the row
				if (*mask)
				{
					// check whether upper, left, lower, and bottom neighbours are in given mask
					if (*(mask - sx) && *(mask - 1) && *(mask + sx) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next row
				index += 3; ptr += 3; ++mask;
			}

			// check the lower left corner of the first slice
			if (*mask)
			{
				// check whether upper, right, and bottom neighbours are in given mask
				if (*(mask - sx) && *(mask + 1) && *(mask + slice))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next grid point
			++index; ++ptr; ++mask;

			// check the remaining grid points in the last row of the first slice except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				if (*mask)
				{
					// check whether upper, left, right, and bottom neighbours are in given mask
					if (*(mask - sx) && *(mask - 1) && *(mask + 1) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;
			}

			// check the lower right corner of the first slice
			if (*mask)
			{
				// check whether upper, left, and bottom neighbours are in given mask
				if (*(mask - sx) && *(mask - 1) && *(mask + slice))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next slice
			index += shift + 3; ptr += shift + 3; ++mask;

			// check the remaining slices except the last one
			for (Type::Size z = 2; z < sz; ++z)
			{
				// check the upper left corner
				if (*mask)
				{
					// check whether top, right, lower, and bottom neighbours are in given mask
					if (*(mask - slice) && *(mask + 1) && *(mask + sx) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;

				// check the remaining grid points in the first row except the last one
				for (Type::Size x = 2; x < sx; ++x)
				{
					if (*mask)
					{
						// check whether top, left, right, lower, and bottom neighbours are in given mask
						if (*(mask - slice) && *(mask - 1) && *(mask + 1) && *(mask + sx) && *(mask + slice))
						{
							*ptr = -1;
						}
						else
						{
							*ptr = 0;
							initial_interface.push_back(index);
						}

						// add a grid point to the foreground
						m_obj_manager.AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager.AddBackground(index);
					}

					// move to the next grid point
					++index; ++ptr; ++mask;
				}

				// check the upper right corner
				if (*mask)
				{
					// check whether top, left, lower, and bottom neighbours are in given mask
					if (*(mask - slice) && *(mask - 1) && *(mask + sx) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next row
				index += 3; ptr += 3; ++mask;

				// check the remaining rows except the last one
				for (Type::Size y = 2; y < sy; ++y)
				{
					// check the first grid point in the row
					if (*mask)
					{
						// check whether top, upper, right, lower, and bottom neighbours are in given mask
						if (*(mask - slice) && *(mask - sx) && *(mask + 1) && *(mask + sx) && *(mask + slice))
						{
							*ptr = -1;
						}
						else
						{
							*ptr = 0;
							initial_interface.push_back(index);
						}

						// add a grid point to the foreground
						m_obj_manager.AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager.AddBackground(index);
					}

					// move to the next grid point
					++index; ++ptr; ++mask;

					// chech the remaining grid points in the row except the last one
					for (Type::Size x = 2; x < sx; ++x)
					{
						if (*mask)
						{
							// check whether all six neighbours are in given mask
							if (*(mask - slice) && *(mask - sx) && *(mask - 1) && *(mask + 1) && *(mask + sx) && *(mask + slice))
							{
								*ptr = -1;
							}
							else
							{
								*ptr = 0;
								initial_interface.push_back(index);
							}

							// add a grid point to the foreground
							m_obj_manager.AddForeground(index);
						}
						else
						{
							*ptr = 1;

							// add a grid point to the background
							m_obj_manager.AddBackground(index);
						}

						// move to the next grid point
						++index; ++ptr; ++mask;
					}

					// check the last grid point in the row
					if (*mask)
					{
						// check whether top, upper, left, lower, and bottom neighbours are in given mask
						if (*(mask - slice) && *(mask - sx) && *(mask - 1) && *(mask + sx) && *(mask + slice))
						{
							*ptr = -1;
						}
						else
						{
							*ptr = 0;
							initial_interface.push_back(index);
						}

						// add a grid point to the foreground
						m_obj_manager.AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager.AddBackground(index);
					}

					// move to the next row
					index += 3; ptr += 3; ++mask;
				}

				// check the lower left corner
				if (*mask)
				{
					// check whether top, upper, right, and bottom neighbours are in given mask
					if (*(mask - slice) && *(mask - sx) && *(mask + 1) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;

				// check the remaining grid points in the last row except the last one
				for (Type::Size x = 2; x < sx; ++x)
				{
					if (*mask)
					{
						// check whether top, upper, left, right, and bottom neighbours are in given mask
						if (*(mask - slice) && *(mask - sx) && *(mask - 1) && *(mask + 1) && *(mask + slice))
						{
							*ptr = -1;
						}
						else
						{
							*ptr = 0;
							initial_interface.push_back(index);
						}

						// add a grid point to the foreground
						m_obj_manager.AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager.AddBackground(index);
					}

					// move to the next grid point
					++index; ++ptr; ++mask;
				}

				// check the lower right corner
				if (*mask)
				{
					// check whether top, upper, left, and bottom neighbours are in given mask
					if (*(mask - slice) && *(mask - sx) && *(mask - 1) && *(mask + slice))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next slice
				index += shift + 3; ptr += shift + 3; ++mask;
			}

			// check the last slice

			// check the upper left corner of the last slice
			if (*mask)
			{
				// check whether top, right, and lower neighbours are in given mask
				if (*(mask - slice) && *(mask + 1) && *(mask + sx))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next grid point
			++index; ++ptr; ++mask;

			// check the remaining grid points in the first row of the last slice except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				if (*mask)
				{
					// check whether top, left, right, and lower neighbours are in given mask
					if (*(mask - slice) && *(mask - 1) && *(mask + 1) && *(mask + sx))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;
			}

			// check the upper right corner of the last slice
			if (*mask)
			{
				// check whether top, left, and lower neighbours are in given mask
				if (*(mask - slice) && *(mask - 1) && *(mask + sx))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next row
			index += 3; ptr += 3; ++mask;

			// check the remaining rows of the last slice except the last one
			for (Type::Size y = 2; y < sy; ++y)
			{
				// check the first grid point in the row
				if (*mask)
				{
					// check whether top, upper, right, and lower neighbours are in given mask
					if (*(mask - slice) && *(mask - sx) && *(mask + 1) && *(mask + sx))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;

				// chech the remaining grid points in the row except the last one
				for (Type::Size x = 2; x < sx; ++x)
				{
					if (*mask)
					{
						// check whether top, upper, left, right, and lower neighbours are in given mask
						if (*(mask - slice) && *(mask - sx) && *(mask - 1) && *(mask + 1) && *(mask + sx))
						{
							*ptr = -1;
						}
						else
						{
							*ptr = 0;
							initial_interface.push_back(index);
						}

						// add a grid point to the foreground
						m_obj_manager.AddForeground(index);
					}
					else
					{
						*ptr = 1;

						// add a grid point to the background
						m_obj_manager.AddBackground(index);
					}

					// move to the next grid point
					++index; ++ptr; ++mask;
				}

				// check the last grid point in the row
				if (*mask)
				{
					// check whether top, upper, left, and lower neighbours are in given mask
					if (*(mask - slice) && *(mask - sx) && *(mask - 1) && *(mask + sx))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next row
				index += 3; ptr += 3; ++mask;
			}

			// check the lower left corner of the last slice
			if (*mask)
			{
				// check whether top, upper, and right neighbours are in given mask
				if (*(mask - slice) && *(mask - sx) && *(mask + 1))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}

			// move to the next grid point
			++index; ++ptr; ++mask;

			// check the remaining grid points in the last row of the first slice except the last one
			for (Type::Size x = 2; x < sx; ++x)
			{
				if (*mask)
				{
					// check whether top, upper, left, and right neighbours are in given mask
					if (*(mask - slice) && *(mask - sx) && *(mask - 1) && *(mask + 1))
					{
						*ptr = -1;
					}
					else
					{
						*ptr = 0;
						initial_interface.push_back(index);
					}

					// add a grid point to the foreground
					m_obj_manager.AddForeground(index);
				}
				else
				{
					*ptr = 1;

					// add a grid point to the background
					m_obj_manager.AddBackground(index);
				}

				// move to the next grid point
				++index; ++ptr; ++mask;
			}

			// check the lower right corner of the last slice
			if (*mask)
			{
				// check whether top, upper, and left neighbours are in given mask
				if (*(mask - slice) && *(mask - sx) && *(mask - 1))
				{
					*ptr = -1;
				}
				else
				{
					*ptr = 0;
					initial_interface.push_back(index);
				}

				// add a grid point to the foreground
				m_obj_manager.AddForeground(index);
			}
			else
			{
				*ptr = 1;

				// add a grid point to the background
				m_obj_manager.AddBackground(index);
			}
		}

		/**********************************************************************/

	} // namespace fls

} // namespace i3d

