/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file se.cc
  \brief Structuring elements - implementation.

  Definition of various types of structuring elements 
 
  \author Petr Matula (pem@fi.muni.cz) 2009
  \author Jiri Kus (207578@mail.muni.cz) 2009
 
*/

#include "se.h"
#include <cmath>
#ifndef PI
	#define PI 3.1415926535897932384626433832795
#endif

using namespace i3d;

namespace se {
/*******GLOBAL FUNCTIONS********/
void printNeighbours(std::ostream &os, const std::vector<Vector3d<int> >nhood){
	os<<"Neighborhood: "<<nhood.size()<<" neighbours"<<std::endl;
	if(nhood.size()>0){
		int xMax,yMax,xMin,yMin;
		xMax=xMin=nhood[0].x;
		yMax=yMin=nhood[0].y;
		for(int i=0;i<(int)nhood.size();i++){
			if(nhood[i].x>xMax){
				xMax=nhood[i].x;
			}else if(nhood[i].x<xMin){
				xMin=nhood[i].x;
			}

			if(nhood[i].y>yMax){
				yMax=nhood[i].y;
			}else if(nhood[i].y<yMin){
				yMin=nhood[i].y;
			}
			os<<nhood[i];
		}
		//contains origin[0,0]?
		//if not,repair values to contain it

		// test x
		if(xMax<0){
			xMax=0;
		}else if(xMin>0){
			xMin=0;
		}//else nothing

		//test y
		if(yMax<0){
			yMax=0;
		}else if(yMin>0){
			yMin=0;
		}//else nothing

		//init
		std::vector<std::vector<bool> > matrix(yMax-yMin+1); //row count
		for(int j=0;j<yMax-yMin+1;j++){
			for(int i=0;i<xMax-xMin+1;i++){
				matrix[j].push_back(false);
			}
		}
		//origin
		int oX= 0-xMin;
		int oY= 0-yMin;
		//load values
		for(int n=0;n<(int)nhood.size();n++){
			int x= nhood[n].x-xMin;
			int y= nhood[n].y-yMin;
			matrix[y][x]=true;
		}
		//print
		os<<std::endl;
		for(int j=0;j<yMax-yMin+1;j++){
			for(int i=0;i<xMax-xMin+1;i++){
				if(i==oX && j==oY){
					os<<(matrix[j][i]?"(x)":"(.)");
				}else{
					os<<(matrix[j][i]?" x ":" . ");
				}
			}
			os<<std::endl;
		}
		
	}
	os<<std::endl;
}

void printSequence(std::ostream &os, const std::vector<StructuringElement*>ses){
	os<<"Decomposition:"<<ses.size()<<" structuring elements"<<std::endl;
	for(int i=0;i<(int)ses.size();i++){
		std::vector<Vector3d<int> >nhood;
		ses[i]->GetNeighbours(nhood);
		printNeighbours(os,nhood);
	}
}

double seround(double x){
	if(x>=0){
		return floor(x+0.5);
	}else{
		return ceil(x-0.5);
	}
}

bool contains(const Vector3d<int>v,const std::vector<Vector3d<int> >nhood){
	for(size_t i=0;i<nhood.size();i++){
		if(v==nhood[i]){
			return true;
		}
	}
	return false;
}

void simpleDilate(const std::vector<Vector3d<int> >in,const Arbitrary se,std::vector<Vector3d<int> >&out){
	std::vector<Vector3d<int> >nhood;
	
	for(size_t i=0;i<in.size();i++){
		out.push_back(in[i]);
		Arbitrary copy = Arbitrary(se);
		copy.Translate(in[i]);
		copy.GetNeighbours(nhood);
		for(size_t j=0;j<nhood.size();j++){
			if(!contains(nhood[j],in)){
				out.push_back(nhood[j]);
			}
		}
	}
}


/********ARBITRARY*********/	
					
void Arbitrary::GetSequence(std::vector<StructuringElement*>&out)const{
	//return itself
	out.clear();
	Arbitrary a(*this);
	Arbitrary * b=new Arbitrary(a);
	out.push_back(b);
}

void Arbitrary::Reflect(){
	for(int i=0;i<(int)nhood.size();i++){
		nhood[i] = -nhood[i];
	}
}

void Arbitrary::Translate(const Vector3d<int> vec){
	for(int i=0;i<(int)nhood.size();i++){
		nhood[i] = nhood[i]+vec;
	}
}

std::ostream& operator<<(std::ostream &os, const Arbitrary &se){

	os<<"Arbitrary: ";
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	return os;
}

/********NON-FLAT ARBITRARY*********/
		
NonFlatArbitrary::NonFlatArbitrary(const NonFlatArbitrary &se){
	std::vector<Vector3d<int> > temp;
	se.GetNeighbours(temp);
	SetNeighbours(temp);
	se.GetWeights(weights);
}

NonFlatArbitrary::NonFlatArbitrary(const std::vector<Vector3d<int> > &nh, const std::vector<double> ws){
	if(ws.size()!=nh.size()){
		throw InternalException("se::NonFlatArbitrary: size of the neigbourhood must equal size of the weights");	
	}
	SetNeighbours(nh);
	weights=ws;
} 
void NonFlatArbitrary::GetSequence(std::vector<StructuringElement*>&out) const{
	//return itself as arbitrary
	out.clear();
	NonFlatArbitrary a(*this);
	Arbitrary * b = new Arbitrary((Arbitrary)a);
	out.push_back(b);

}
std::ostream& operator<<(std::ostream &os, const NonFlatArbitrary &se){
	std::vector<Vector3d<int> > temp;
	se.GetNeighbours(temp);

	std::vector<double> wTemp;
	se.GetWeights(wTemp);
	os<<"NonFlatArbitrary: ";
	for(int i=0;i<(int)temp.size();i++){
		os<<temp[i]<<":"<<wTemp[i]<<" ";
	}
	return os;
}		
/*********DISK***********/	

Disk::Disk(){
	radius=1;
	SetN(0);
	shift = Vector3d<int>(0,0,0);

}

Disk::Disk(const Disk &se){
	radius=se.GetRadius();
	SetN(se.GetN());
	shift=se.GetShift();
}

Disk::Disk(unsigned r, unsigned nn){
	radius=r;
	SetN(nn);
	shift = Vector3d<int>(0,0,0);
}

void Disk::SetN(unsigned nn){
			if(nn!=0 && nn!=4 && nn!=6 && nn!=8){
				throw InternalException("se::Disk: n must be 0, 4, 6, or 8");
			}else{
				if(radius<3){
					n=0;
				}else{
					n=nn;
				}
			}
		}

void Disk::GetNeighbours(std::vector<Vector3d<int> >&out) const{
	out.clear();
	//traverse the matrix surrounding the disk from bottom to top (and from left to right)
	//add points within the disk to the vector out
	if(n==0){
		for(int y=-(int)(radius);y<=(int)(radius);y++){
			for(int x=-(int)(radius);x<=(int)(radius);x++){
				if((y*y+x*x)<=(int)(radius*radius)){
					out.push_back(Vector3d<int>(x,y,0)+shift);
				}
			}
		}
	}else{
		//dilate sequence decomposing the SE and get neighbours from it
		std::vector<StructuringElement*>sequence;
		GetSequence(sequence);
		out.clear();
		out.push_back(Vector3d<int>(0,0,0)+shift);
		Arbitrary a;
		for(size_t i=0;i<sequence.size();i++){
			std::vector<Vector3d<int> >in;
			sequence[i]->GetNeighbours(in);
			a=Arbitrary(in);
			std::vector<Vector3d<int> > result;
			simpleDilate(out,a,result);
			out=result;
		}
		//free memory
		for(size_t i=0;i<sequence.size();i++){
			delete sequence[i];
		}
	}
}
void Disk::GetSequence(std::vector<StructuringElement*>&out) const{
    /*Reference for radial decomposition of disks:  Rolf Adams, "Radial
	% Decomposition of Discs and Spheres," CVGIP:  Graphical Models and
	% Image Processing, vol. 55, no. 5, September 1993, pp. 325-332.
	
	% The specific decomposition technique used here is radial
	% decomposition using periodic lines.  The reference is:  Ronald
	% Jones and Pierre Soille, "Periodic lines: Definition, cascades, and
	% application to granulometries," Pattern Recognition Letters,
	% vol. 17, 1996, pp. 1057-1063.
	
	  comments contains related fragments of code from matlab
	*/
	std::vector<int> v;
	out.clear();

	switch(n){ 
	  case 0:
		  {
		  	//return itself
			out.clear();
			Disk a(*this);
			Disk * b = new Disk(a);
			out.push_back(b);
			return;
		  }
	  case 4:
		  {
			int arr[] = {0,1,
						 1,1,
						 1,0,
						 1,-1};
			v = std::vector<int>(arr, arr + sizeof(arr) / sizeof(int) );
		  }
		 break;	
	  case 6:
		  {
		  int arr[] = {0,1,
					   2,1,
					   1,2,
					   1,0,
			   		   2,-1,
					   1,-2};
		  v = std::vector<int>(arr, arr + sizeof(arr) / sizeof(int) );
		  break;
		  }
	  case 8:
		  {
		  int arr[] = {0,1,
					   1,2,
				       1,1,
				       2,1,
			           1,0,
			           2,-1,
			           1,-1,
					   1,-2};
		  v = std::vector<int>(arr, arr + sizeof(arr) / sizeof(int) );
		  break;
		  }	
	  default: 
        //% This error should have been caught already in Constructor;
		throw InternalException("se::Disk: n must be 0, 4, 6, or 8");
		break;
	}
	
    /* Determine k, which is the desired radial extent of the periodic
       line strels.  For the origin of this formula, see the second
       paragraph on page 328 of the Rolf Adams paper.*/
    double theta = PI/(2*n);
    double k = (double)2*radius/((double)1/tan(theta) + (double)1/sin(theta));

    /*For each periodic line strel, determine the repetition parameter,
      rp.  The use of floor() in the computation means that the resulting
      strel will be a little small, but we will compensate for this
      below.*/
	for(int q=0;q<(int)n;q++){ //for q = 1:n
		// matlab:
		// NORM(V,P) = sum(abs(V).^P)^(1/P).
		// NORM(V) = norm(V,2).
		double norm = sqrt((double)abs(v[2*q])*abs(v[2*q])+(abs(v[2*q+1])*abs(v[2*q+1]))); // norm(v(q,:))
        int rp = (int)floor(k/norm);
		//se.decomposition(q) = strel('periodicline', rp, v(q,:));
		PeriodicLine * ap = new PeriodicLine(rp,v[2*q],v[2*q+1]);
		out.push_back(ap);
	}

    /* Now dilate the strels in the decomposition together to see how
      close we came to the desired disk radius.*/

    //nhood = imdilate(1, se.decomposition, 'full');
	std::vector<Vector3d<int> > dilatedNhood;
	dilatedNhood.push_back(Vector3d<int>(0,0,0));
	Arbitrary a;
	for(size_t i=0;i<out.size();i++){
		std::vector<Vector3d<int> >in;
		out[i]->GetNeighbours(in);
		a=Arbitrary(in);
		std::vector<Vector3d<int> > result;
		simpleDilate(dilatedNhood,a,result);
		dilatedNhood=result;
	}

    /*nhood = nhood > 0;
    [rd,cd] = find(nhood);
    M = size(nhood,1);
    rd = rd - floor((M+1)/2);
    max_horiz_radius = max(rd(:));
    radial_difference = r - max_horiz_radius;
	*/
	int xMax=0;
		for(int i=0;i<(int)dilatedNhood.size();i++){
			if(dilatedNhood[i].x>xMax){
				xMax=dilatedNhood[i].x;
			}
		}
	int radial_difference = radius-xMax;

    /* Now we are going to add additional vertical and horizontal line
      strels to compensate for the fact that the strel resulting from the
      above decomposition tends to be smaller than the desired size.*/
    int len = 2*(radial_difference-1) + 1;
	if (len >= 3){
        //% Add horizonal and vertical line strels.
        Line* line=new Line(len,0);
		out.push_back(line);//se.decomposition(end+1) = strel('line',len,0);
		line=new Line(len,90);
        out.push_back(line);//se.decomposition(end+1) = strel('line',len,90);
	}
	//translate first se of decomposition
	out[0]->Translate(shift);
}


std::ostream& operator<<(std::ostream &os, const Disk &se){

	os<<"Disk ("<<se.GetRadius()<<","<<se.GetN()<<"), "<<std::endl;

	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	std::vector<StructuringElement*> sequence;
	se.GetSequence(sequence);
	printSequence(os,sequence);
	return os;
}

/*******BALL*******/

Ball::Ball(){
	SetRadius(1);
	SetN(4);
	weight=1.0;
	SetShift(Vector3d<int>(0,0,0));
}

Ball::Ball(const Ball &se){
	weight=se.GetWeight();
	SetRadius(se.GetRadius());
	SetN(se.GetN());
	SetShift(se.GetShift());

}



Ball::Ball(unsigned radius, double w,unsigned n){
	SetRadius(radius);
	SetN(n);
	SetShift(Vector3d<int>(0,0,0));
	weight=w;

}

void Ball::GetWeights(std::vector<double>&out) const{
	//se.height = h * sqrt(r^2 - min(r^2,xx.^2 + yy.^2)) / r;
		out.clear();
		int r = GetRadius();
	//traverse the matrix surrounding the disk from bottom to top (and from left to right)
	//and set weights
	for(int y=-(int)(r);y<=(int)(r);y++){
		for(int x=-(int)(r);x<=(int)(r);x++){
			if((y*y+x*x)<=(r*r)){	
				double height = weight * sqrt((double)(r*r - std::min(r*r,y*y + x*x))) / r;
				out.push_back(height);
			}
		}
	}
}

std::ostream& operator<<(std::ostream &os, const Ball &se){

	os<<"Ball ("<<se.GetRadius()<<","<<se.GetN()<<")"<<std::endl;
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	os<<std::endl;
	std::vector<double> weights;
	se.GetWeights(weights);
	for(int i=0;i<(int)weights.size();i++){
		os<<weights[i]<<", ";
	}
	return os;
}


/*******DIAMOND********/	

Diamond::Diamond(){
	radius=1;
	shift = Vector3d<int>(0,0,0);
}

Diamond::Diamond(const Diamond &se){
	radius=se.GetRadius();
	shift=se.GetShift();
}

Diamond::Diamond(const unsigned r){
	radius = r;
	shift = Vector3d<int>(0,0,0);
}

void Diamond::GetNeighbours(std::vector<Vector3d<int> >&out) const{
	//size of rectangle containing diamond minus corners
	// size of corner = radius*(radius+1)/2
	int size= radius*2+1;
	int corner= ((int)radius*(radius+1))/2;
	out.resize(size*size-4*corner);
	int i=0;
	//rows
	for(int y=-(int)(radius);y<=(int)(radius);y++){
		//columns
		for(int x=-(int)(radius);x<=(int)(radius);x++){
			//is part of diamond?
			if((abs(x)+abs(y))<=(int)radius){
				out[i]=(Vector3d<int>(x,y,0)+shift);
				i++;
			}
		}
	}
}
void Diamond::GetSequence(std::vector<StructuringElement*>&out) const{
	out.clear();
	/*Heuristic --- if M > 2, assume computational advantage of decomposition
	  is worth the cost of performing multiple dilations (or erosions).*/
  if (radius > 2){
	/*Compute the logarithmic decomposition of the strel using the method in
	 Rein van den Boomgard and Richard van Balen, "Methods for Fast
	 Morphological Image Transforms Using Bitmapped Binary Images," CVGIP:
	 Models and Image Processing, vol. 54, no. 3, May 1992, pp. 252-254.*/
	float number = log((float)radius);
	float base = log((float)2);
	int n = (int)floor(number/base);

	std::vector<Vector3d<int> >nhood; 
	
	nhood.push_back(Vector3d<int>(0,-1,0));
	nhood.push_back(Vector3d<int>(-1,0,0));
	nhood.push_back(Vector3d<int>(0,0,0));
	nhood.push_back(Vector3d<int>(1,0,0));
	nhood.push_back(Vector3d<int>(0,1,0));
	Arbitrary * ap = new Arbitrary(nhood);
	ap->Translate(shift);
	out.push_back(ap); //se.decomposition = strel([0 1 0; 1 1 1; 0 1 0]);
	
	for(int k=0;k<n;k++){ // k = 0:(n-1)
		int P = (int)floor(pow((float)2,(k+1))+0.5) + 1;
		int middle = (P+1)/2;
		nhood.clear();//nhood = zeros(P,P);
		nhood.push_back(Vector3d<int>(1,middle,0));//nhood(1,middle) = 1;
		nhood.push_back(Vector3d<int>(P,middle,0));//nhood(P,middle) = 1;
		nhood.push_back(Vector3d<int>(middle,1,0));//nhood(middle,1) = 1;
		nhood.push_back(Vector3d<int>(middle,P,0));//nhood(middle,P) = 1; 
		ap = new Arbitrary(nhood);
		ap->Translate(Vector3d<int>(-middle,-middle,0));
		out.push_back(ap); //se.decomposition(end+1) = strel(nhood);
	}
	int q = (int)radius - (int)floor(pow((float)2,n)+0.5);
	if (q > 0){
	   int P = 2*q+1;
	   int middle = (P+1)/2;
	   nhood.clear();//nhood = zeros(P,P);
       nhood.push_back(Vector3d<int>(1,middle,0));//nhood(1,middle) = 1;
       nhood.push_back(Vector3d<int>(P,middle,0));//nhood(P,middle) = 1;
       nhood.push_back(Vector3d<int>(middle,1,0));//nhood(middle,1) = 1;
       nhood.push_back(Vector3d<int>(middle,P,0));//nhood(middle,P) = 1;
       ap = new Arbitrary(nhood);
	   ap->Translate(Vector3d<int>(-middle,-middle,0));
	   out.push_back(ap); //se.decomposition(end+1) = strel(nhood);
    }
  }else{
	 //return itself
	 Diamond a(*this);
	 Diamond * ap = new Diamond(a);
	 out.push_back(ap);
  }
}

std::ostream& operator<<(std::ostream &os, const Diamond &se){
	
	os<<"Diamond("<<se.GetRadius()<<"): ";
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	std::vector<StructuringElement*> sequence;
	se.GetSequence(sequence);
	printSequence(os,sequence);
	return os;
}	



/********LINE*********/	

Line::Line(){
	length=0; 
	angle=0; 
	shift=Vector3d<int>(0,0,0);
	sign=Vector3d<int>(1,1,1);
}

Line::Line(const Line &se){
	angle=se.GetAngle();
	length=se.GetLength();
	shift=se.GetShift();
	sign=se.GetSign();
}

Line::Line(unsigned l, unsigned a){
	length=l;
	angle=a;
	shift=Vector3d<int>(0,0,0);
	sign=Vector3d<int>(1,1,1);
}

void Line::GetNeighbours(std::vector<Vector3d<int> >&out)const{
	out.clear();
	if (length >= 1){
	 
	 //find point
	 double theta = angle * PI / 180;
	 int x =  (int)seround((((double)length-1)/2 * cos(theta)));
	 int y =  (int)-seround((((double)length-1)/2 * sin(theta)));
	
	 int x1 =-x;
	 int y1 =-y;
	 int x2 =x;
	 int y2 =y;

	//intline
	int dx = abs(x2 - x1);
	int dy = abs(y2 - y1);

	// Check for degenerate case.
	if ((dx == 0) && (dy == 0)){
		out.push_back(sign*(Vector3d<int>(x1,y1,0))+shift);
	  return;
	}

	if (dx >= dy){
		if (x1 > x2){
		// Always "draw" from left to right.
		int t;
		t = x1; x1 = x2; x2 = t;
		t = y1; y1 = y2; y2 = t;
	  }
	  double m = (double)(y2 - y1)/(double)(x2 - x1);

	  for(int i=x1;i<=x2;i++){
		out.push_back(sign*(Vector3d<int>(i,(int)seround((y1 + m*(i - x1))),0))+shift);
	  }
	  
	}else{
		if (y1 > y2){
		// Always "draw" from bottom to top.
		int t;
		t = x1; x1 = x2; x2 = t;
		t = y1; y1 = y2; y2 = t;
		}
		double m = (double)(x2 - x1)/(double)(y2 - y1);

		  for(int i=y1;i<=y2;i++){
			out.push_back(sign*(Vector3d<int>((int)seround((x1 + m*(i - y1))),i,0))+shift);
		  }
	    }
	}//if length<1 do nothing
}
void Line::GetSequence(std::vector<StructuringElement*>&out) const{
	//return itself
	Line a(*this);
	Line * ap = new Line(a);
	out.push_back(ap);

}

std::ostream& operator<<(std::ostream &os, const Line &se){

	os<<"Line("<<se.GetLength()<<","<<se.GetAngle()<<"): "<<std::endl;
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	return os;
}	


/*******PAIR********/

/// Copy constructor.
Pair::Pair(const Pair &se){
	first=se.GetFirst();
	second=se.GetSecond();
}

/** Creates a flat structuring element containing two members. 
	One member is located at the origin. 
	The second is lcoated at [a,b,c] */
Pair::Pair(int a, int b, int c){
	first=Vector3d<int>(0,0,0);
	second=Vector3d<int>(a,b,c); 
}
Pair::Pair(const Vector3d<int> vec){
	first=Vector3d<int>(0,0,0);
	second=vec;
}

//StructuringElement functions
void Pair::GetNeighbours(std::vector<Vector3d<int> >&out) const{
	out.resize(2);
	out[0]=(first);
	out[1]=(second);
}

void Pair::GetSequence(std::vector<StructuringElement*>&out) const{
	//return itself
	out.clear();
	Pair a(*this);
	Pair * ap = (new Pair(a));
	out.push_back(ap);
}

void Pair::Reflect(){
	first = -first;
	second = -second;
}

void Pair::Translate(const Vector3d<int>vec){
	first += vec;
	second += vec;
}

	std::ostream& operator<<(std::ostream &os, const Pair &se){
		os<<"Pair: "<<std::endl;
			std::vector<Vector3d<int> > neighbours;
			se.GetNeighbours(neighbours);
			printNeighbours(os,neighbours);
		return os;
}


/********PERIODIC LINE************/

PeriodicLine::PeriodicLine(){
	even=false;
	p=0;
	rowOffset=0;
	colOffset=0;
	shift = Vector3d<int>(0,0,0);
	sign = Vector3d<int>(1,1,1);
}

PeriodicLine::PeriodicLine(const PeriodicLine &se){
	if(se.IsEven()){
		even=true;
		p=se.GetP();
		n=se.GetN();
	}else{
		even=false;
		p=se.GetP();
	}
	rowOffset=se.GetRowOffset();
	colOffset=se.GetColOffset();
	shift=se.GetShift();
	sign = GetSign();
}

PeriodicLine::PeriodicLine(int pe,int en, int co,int ro){
	if(pe<0 || en<0){
		throw InternalException("se::PeriodicLine: p and n must be non-negative");
	}
		even=true;
		p=pe;
		n=en;
		rowOffset=ro;
		colOffset=co;
		shift=Vector3d<int>(0,0,0);
		sign = Vector3d<int>(1,1,1);
}

PeriodicLine::PeriodicLine(int pe, int co, int ro){
	if(pe<0){
		throw InternalException("se::PeriodicLine: p must be non-negative");
	}
	even=false;
	p=pe;
	rowOffset=ro;
	colOffset=co;
	shift = Vector3d<int>(0,0,0);
	sign = Vector3d<int>(1,1,1);
}

void PeriodicLine::GetNeighbours(std::vector<Vector3d<int> >&out) const{
	if(even){
		out.resize(p);
		int index=0;
		for(int i=(0-n);i<(p-n);i++){
			out[index]=(sign*(Vector3d<int>(i*colOffset,i*rowOffset,0))+shift);
			index++;
		}
	}else{
		out.resize(p*2+1);
		out[0]=(Vector3d<int>(0,0,0)+shift);
		int index=1;

		for(int i=1;i<=p;i++){
			out[index]=(Vector3d<int>(i*colOffset,i*rowOffset,0)+shift);
			out[index+1]=(Vector3d<int>(-i*colOffset,-i*rowOffset,0)+shift);
			index+=2;
		}
	}
}

void PeriodicLine::GetSequence(std::vector<StructuringElement*>&out) const{
	//return itself
	PeriodicLine a(*this);
	PeriodicLine* ap = (new PeriodicLine(a));
	out.push_back(ap);
}

std::ostream& operator<<(std::ostream &os, const PeriodicLine &se){
	
	os<<"PeriodicLine("<<se.GetP()<<","<<se.GetColOffset()<<","<<se.GetRowOffset()<<"): "<<std::endl;
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	return os;
}	


/***********RECTANGLE*****************/

Rectangle::Rectangle(){
	rows = 3;
	cols = 3;
	shift=Vector3d<int>(-1,-1,0);
	sign = Vector3d<int>(1,1,1);
}

Rectangle::Rectangle(const Rectangle &se){
	cols=se.GetCols();
	rows=se.GetRows();
	shift=se.GetShift();
	sign=se.GetSign();
}

Rectangle::Rectangle(unsigned c, unsigned r){
	cols=c; rows=r;
	shift=Vector3d<int>(-(int)(cols/2),-(int)(rows/2),0);
	sign = Vector3d<int>(1,1,1);
}
void Rectangle::Reflect(){
	sign = -sign;
	shift= -shift;
}
void Rectangle::GetNeighbours(std::vector<Vector3d<int> >&out) const{
	out.resize(cols*rows);
	int index=0;
	for(int x=0;x<(int)cols;x++){
		for(int y=0;y<(int)rows;y++){
			out[index]=(sign*(Vector3d<int>(x,y,0))+shift);
			index++;
		}
	}
}
void Rectangle::GetSequence(std::vector<StructuringElement*>&out) const{
	//do not decompose rectangles 
	if((GetCols()>1) & (GetRows()>1)){
		std::vector<Vector3d<int> > horizontal;
		std::vector<Vector3d<int> > vertical;
		for(int i=0;i<(int)cols;i++){
			horizontal.push_back(Vector3d<int>(-(int)cols/2+i,0,0));
		}
		for(int i=0;i<(int)rows;i++){
			vertical.push_back(Vector3d<int>(0,-(int)rows/2+i,0));
		}
		Arbitrary * ap = new Arbitrary(horizontal);
		//shift first SE if necessary
		if(sign.x>0){
			ap->Translate(shift);
		}else{
			ap->Reflect();
			ap->Translate(sign*shift);
		}

		out.push_back(ap);
		ap =new Arbitrary(vertical);
		out.push_back(ap);
	}else{
		//return itself
		Rectangle r = Rectangle(*this);

		Rectangle * ap = new Rectangle(r);
		out.push_back(ap);
	}

}

std::ostream& operator<<(std::ostream &os, const Rectangle &se){

	os<<"Rectangle("<<se.GetCols()<<","<<se.GetRows()<<"): "<<std::endl;
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	std::vector<StructuringElement*> sequence;
	se.GetSequence(sequence);
	printSequence(os,sequence);

	return os;
}		



/***********SQUARE************/
Square::Square(){
	SetRows(3);
	SetCols(3);
	SetShift(Vector3d<int>(-1,-1,-1));
	SetSign(Vector3d<int>(1,1,1));
}

Square::Square(const Square &se){
	SetRows(se.GetCols());
	SetCols(se.GetRows());
	SetShift(se.GetShift());
	SetSign(se.GetSign());
}

Square::Square(unsigned s){
	SetRows(s);
	SetCols(s);
	SetShift(Vector3d<int>(-(int)(s/2),-(int)(s/2),0));
	SetSign(Vector3d<int>(1,1,1));
}



std::ostream& operator<<(std::ostream &os, const Square &se){
	
	os<<"Square("<<se.GetSize()<<"): "<<std::endl;
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	std::vector<StructuringElement*> sequence;
	se.GetSequence(sequence);
	printSequence(os,sequence);

	return os;
}		

/***********OCTAGON**********/

Octagon::Octagon(){
	radius=0;
	shift = Vector3d<int>(0,0,0);
}

Octagon::Octagon(const Octagon &se){
	radius=se.GetRadius();
	shift=se.GetShift();
}

Octagon::Octagon(int r){
	if(r<0 || r%3!=0){
		throw InternalException("se::Octagon: size of the octagon must be a nonnegative multiple of 3");
	}
	radius=r;
	shift = Vector3d<int>(0,0,0);
}

void Octagon::GetNeighbours(std::vector<Vector3d<int> >&out) const{
	out.clear();
	//rows
	//push_back [0,0]
	out.push_back(Vector3d<int>(0,0,0)+shift);

	int modifier=0;
	int n =radius/3;
	for(int j=0;j<=radius;j++){
		if(j>n){
			modifier++;
		}
		for(int i=1;i<=(radius-modifier);i++){
			//push_back [i,j]
			out.push_back(Vector3d<int>(i,j,0)+shift);
			//push_back 3 symmetric points
			out.push_back(Vector3d<int>(-i,j,0)+shift);
			if(j!=0){
				out.push_back(Vector3d<int>(i,-j,0)+shift);
				out.push_back(Vector3d<int>(-i,-j,0)+shift);
			}else{
				out.push_back(Vector3d<int>(0,-i,0)+shift);
				out.push_back(Vector3d<int>(0,i,0)+shift);
			}	
		}
	}
}
void Octagon::GetSequence(std::vector<StructuringElement*>&out) const{
	out.clear();

	int k = radius/3;
	if (k < 1){
		//single point
		std::vector<Vector3d<int> >nhood;
		nhood.push_back(Vector3d<int>(0,0,0));
		Arbitrary * ap = (new Arbitrary(nhood));
		//shift first SE if necessary
		ap->Translate(shift);
		out.push_back(ap);
	}else{
		//decompose into 4 line strels.
		//a = [0 0 0; 1 1 1; 0 0 0];
		PeriodicLine * ap = (new PeriodicLine(k,0,1));
		//shift first SE if necessary
		ap->Translate(shift);
		out.push_back(ap);//se.decomposition = strel(a);
		//b = a';
		ap=(new PeriodicLine(k,1,0));
		out.push_back(ap);//se.decomposition(2) = strel(b);
		//c = eye(3);
		ap=(new PeriodicLine(k,1,1));
		out.push_back(ap);//se.decomposition(3) = strel(c);
		//d = rot90(c);
		ap=(new PeriodicLine(k,-1,1));
		out.push_back(ap);//se.decomposition(4) = strel(d);	

	}
}

std::ostream& operator<<(std::ostream &os, const Octagon &se){

	os<<"Octagon("<<se.GetRadius()<<"):"<<std::endl;
	std::vector<Vector3d<int> > neighbours;
	se.GetNeighbours(neighbours);
	printNeighbours(os,neighbours);

	std::vector<StructuringElement*> sequence;
	se.GetSequence(sequence);
	printSequence(os,sequence);
	
	return os;
}	

};
