/*
* i3dlib - image manipulation library
*
* Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, write to the Free
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/** \file watershed_topolgical.cc
\brief Part of mathematical morphology routines - implementation.

This file contains bodies of mathematical morphology routines dealing with 
the topological watershed transform.
*/ 

#include "watersheds.h"
#include "component_tree.h"
#include "i3dassert.h"
#include <limits>
#include <stack>
#include <set>
#include <map>
#include <deque>
#ifdef I3D_DEBUG
#include <iostream>
#endif // #ifdef I3D_DEBUG

#ifdef __GNUG__
#	pragma implementation
#endif

namespace i3d {

	//--------------------------------------------------------------------------
	// Topological Watershed transform implementation
	//--------------------------------------------------------------------------

	/** Handler of the additional attributes associated to the component tree node used for the topological watershed computation - depth, first occurance.
	Structure of the attributes after the build process: size_t depth, size_t firstOc; (depth of the component inside the tree, index of the first occurance in the DFS list)
	*/
	template <typename T>
	class I3D_ALGO_EXPORT CCompTreeNodeAttribHandlerTW: public CCompTreeNodeAttribHandler<T>
	{
	public:
		/// constructor that initializes the default attributes
		CCompTreeNodeAttribHandlerTW()
		{
			numOfNodes = 0;
		}

		/// frees the space allocated for the attributes in the node and calls itself recursively on the children
		virtual void DeletingNode(SCompTreeNode<T> *node)
		{
			free(node->attributes);
			node->attributes = NULL;
			for (SCompTreeNodeListElem<T> *i = node->children.first; i != NULL; i = i->next)
				DeletingNode(i->value);
		}
		/// computes the depth for the whole tree
		virtual void BuildingFinished(SCompTreeNode<T> *node)
		{
			// for the root node allocate new space for the new attributes and set them to default values
			if (!node->attributes)
			{
				node->attributes = malloc(2 * sizeof(size_t));
				SetDepth(node, 0);
				SetFirstOc(node, std::numeric_limits<size_t>::max());
			}
			// OK, we are sure that this node has it's atributes set, compute the attributes for all the children
			size_t newDepth = GetDepth(node) + 1;
			for (SCompTreeNodeListElem<T> *i = node->children.first; i != NULL; i = i->next)
			{
				i->value->attributes = malloc(2 * sizeof(size_t));
				SetDepth(i->value, newDepth);
				SetFirstOc(i->value, std::numeric_limits<size_t>::max());
				BuildingFinished(i->value);
			}
			// remember we have processed this node
			++numOfNodes;
		}
		/// prints deph into the output stream
		virtual void PrintAttrib(std::ostream &ost, const SCompTreeNode<T> *node)
		{
			// this should be called only on the node with the final attributes layout, not during the build process!
			std::cout << "depth: " << GetDepth(node) << "; first oc.: " << GetFirstOc(node);
		}

		/// sets depth of the given node
		void SetDepth(SCompTreeNode<T> *node, size_t depth) const
		{
			*(static_cast<size_t*>(node->attributes)) = depth;
		}
		/// returns depth of the given node
		size_t GetDepth(const SCompTreeNode<T> *node) const
		{
			return *(static_cast<size_t*>(node->attributes));
		}
		/// sets first occurance of the given node
		void SetFirstOc(SCompTreeNode<T> *node, size_t firstOc) const
		{
			*(static_cast<size_t*>(node->attributes) + 1) = firstOc;
		}
		/// returns first occurance of the given node
		size_t GetFirstOc(const SCompTreeNode<T> *node) const
		{
			return *(static_cast<size_t*>(node->attributes) + 1);
		}

		/// returns the number of nodes handled by this attribute handler
		size_t GetNumOFNodes() const { return numOfNodes; }

	private:
		/// number of handled nodes
		size_t numOfNodes;
	};

	/// compares two component tree nodes - the one with the lower altitude is the smaller one, in the case of same altitudes, the on without the children is lower (when both the altitude and the fact of having a childre is the same, pointer decides)
	template <typename T>
	struct SCompTreeNodeCompareFunctor
	{
		bool operator()(const SCompTreeNode<T>* node1, const SCompTreeNode<T>* node2)
		{
			if (node1->level < node2->level)
				return true;
			if (node1->level == node2->level)
			{
				if (node1->children.first == NULL && node2->children.first)
					return true;
				return node1 < node2;
			}
			return false;
		}
	};

	/// Representation of the component tree used for the topological watershed computation
	template <typename T>
	struct SCompTreeTW: public SCompTree<T>
	{
		/// constructor that initializes default parameters
		SCompTreeTW(CCompTreeNodeAttribHandlerTW<T> *atribHandler = NULL): SCompTree<T>(atribHandler)
		{
			dFSList = NULL;
			dFSListSize = 0;
		}
		/// destructor that frees all the allocated resources
		virtual ~SCompTreeTW()
		{
			delete[] dFSList;
			for (size_t i = 0; i < dFSListSize; ++i)
				delete[] minInd[i];
			delete[] minInd;
		}

		/// finishes component tree for topological watershed - does the level inversion (tree should be computed from the inverted input image, so the levels should be inverted too) and computes the DFS list and related structures for LCA queries
		void Finish(T maxInputValue)
		{
			// invert the leve value for each componenet, create the DFS list and compute first occurances
			CCompTreeNodeAttribHandlerTW<T>* atribHandlerTW = static_cast<CCompTreeNodeAttribHandlerTW<T>*>(this->atribHandler);
			dFSListSize = atribHandlerTW->GetNumOFNodes() * 2 - 1;
			dFSList = new SCompTreeNode<T>*[dFSListSize];
			std::stack<SCompTreeNode<T>*> dFSStack;
			dFSStack.push(this->root);
			size_t dFSListIndex = 0;
			while (!dFSStack.empty())
			{
				SCompTreeNode<T>* node = dFSStack.top();
				dFSStack.pop();
				if (atribHandlerTW->GetFirstOc(node) > dFSListIndex)
				{
					// we see this node for the first time
					atribHandlerTW->SetFirstOc(node, dFSListIndex);
					node->level = maxInputValue - node->level;
					for (SCompTreeNodeListElem<T> *childNodeListElem = node->children.first; childNodeListElem != NULL; childNodeListElem = childNodeListElem->next)
					{
						dFSStack.push(node);
						dFSStack.push(childNodeListElem->value);
					}
				}
				I3DASSERT_BT(dFSListIndex != dFSListSize, "DFS list size exceeded - list construction error");
				dFSList[dFSListIndex++] = node;
			}
			I3DASSERT_BT(dFSListIndex == dFSListSize, "DFS does not contain 2n + 1 elements");

			// compute the sparse table used for the RMQ queries during the LCA queries (precomputed results of the RMQs of the sizes 2^j; 0 <= j <= \lfloor log_2(dFSListSize) \rfloor)
			// TODO we can save dFSListSize space here if we change this to 1 <= j <= \lfloor log_2(dFSListSize) \rfloor
			// create the sparse table
			size_t maxJ = static_cast<size_t>(floor(log(static_cast<double>(dFSListSize)) / log(2.0)));
			minInd = new size_t*[dFSListSize];
			for (size_t i = 0; i < dFSListSize; ++i)
			{
				for (size_t j = 0; j <= maxJ; ++j)
				{
					minInd[i] = new size_t[maxJ + 1];
					minInd[i][0] = i;
				}
			}
			// compute the values in the sparse table
			for (size_t j = 1; j <= maxJ; ++j)
			{
				for (size_t i = 0; i + (1 << j) - 1 < dFSListSize; ++i)
				{
					if (atribHandlerTW->GetDepth(dFSList[minInd[i][j - 1]]) < atribHandlerTW->GetDepth(dFSList[minInd[i + (1 << (j - 1))][j - 1]]))
						minInd[i][j] = minInd[i][j - 1];
					else
						minInd[i][j] = minInd[i + (1 << (j - 1))][j - 1];
				}
			}
		}

		/// prints whole DFS list into the output stream
		void PrintDFSList(std::ostream &ost) const
		{
			ost << "DFS list:" << std::endl;
			for (size_t i = 0; i < dFSListSize; ++i)
			{
				ost << dFSList[i] << ": ";
				ost << "depth: " << static_cast<CCompTreeNodeAttribHandlerTW<T>*>(this->atribHandler)->GetDepth(dFSList[i]);
				ost << "; first oc.: " << static_cast<CCompTreeNodeAttribHandlerTW<T>*>(this->atribHandler)->GetFirstOc(dFSList[i]);
				ost << std::endl;
			}
		}

		/// returns least common ancestor of the two nodes
		SCompTreeNode<T> *GetLCA(SCompTreeNode<T> *node1, SCompTreeNode<T> *node2) const
		{
			if (node1 == node2)
				return node1;
			// get the interval borders i and j in the dFSListSize for the two components
			CCompTreeNodeAttribHandlerTW<T>* atribHandlerTW = static_cast<CCompTreeNodeAttribHandlerTW<T>*>(this->atribHandler);
			size_t i = atribHandlerTW->GetFirstOc(node1);
			size_t j = atribHandlerTW->GetFirstOc(node2);
			if (j < i)
				std::swap(i, j);
			// compute the size of the computed sub-intervals we want to check
			size_t k = static_cast<size_t>(floor(log(static_cast<double>(j - i + 1)) / log(2.0)));
			// get the lower of the two intervals
			if (atribHandlerTW->GetDepth(dFSList[minInd[i][k]]) <= atribHandlerTW->GetDepth(dFSList[minInd[j - (1 << k) + 1][k]]))
				return dFSList[minInd[i][k]];
			else
				return dFSList[minInd[j - (1 << k) + 1][k]];
		}

		/// returns the highest fork of the set of nodes where the first node is (one of) the lowest element(s)
		SCompTreeNode<T> *GetHighestFork(const std::set<SCompTreeNode<T>*, SCompTreeNodeCompareFunctor<T> > &v) const
		{
			if (v.empty())
				return NULL;
			typename std::set<SCompTreeNode<T>*, SCompTreeNodeCompareFunctor<T> >::const_iterator i = v.begin();
			const SCompTreeNode<T> *minComponent = *i;
			SCompTreeNode<T> *highestForkCandidate = *i;
			typename std::set<SCompTreeNode<T>*, SCompTreeNodeCompareFunctor<T> >::const_iterator iEnd = v.end();
			while (++i != iEnd)
			{
				SCompTreeNode<T> *lca = GetLCA(*i, highestForkCandidate);
				if (lca != *i)
					highestForkCandidate = lca;
			}
			if (highestForkCandidate != minComponent)
				return highestForkCandidate;
			return NULL;
		}

		/// returns the component to which the given voxel can be lowered if it is M-destructible or NULL otherwise
		SCompTreeNode<T> *MDestructible(const Image3d<T> &image, size_t index, NeighbourhoodWalker &nbhWalker) const
		{
			// TODO do we need an image here? when the tree is not simplified, probably no, but what about the situation when we simplify the tree?? 
			// find the components of the lower neighbours
			std::set<SCompTreeNode<T>*, SCompTreeNodeCompareFunctor<T> > lowerNeighbourComponents;
			T voxelHeight;
			GetLowerNeighbourComponents(image, index, lowerNeighbourComponents, voxelHeight, nbhWalker);
			// decide M-Destructability for points with some lower neighbours
			if (lowerNeighbourComponents.empty())
				return NULL;
			SCompTreeNode<T> *minComponent = *lowerNeighbourComponents.begin();
			if (minComponent->children.first)
				return NULL;
			if (!GetHighestFork(lowerNeighbourComponents))
				return minComponent;
			return NULL;
		}

		/// returns the component to which the given voxel can be lowered if it is W-destructible or NULL otherwise
		SCompTreeNode<T> *WDestructible(const Image3d<T> &image, size_t index, NeighbourhoodWalker &nbhWalker) const
		{
			// find the components of the lower neighbours
			std::set<SCompTreeNode<T>*, SCompTreeNodeCompareFunctor<T> > lowerNeighbourComponents;
			T voxelHeight;
			GetLowerNeighbourComponents(image, index, lowerNeighbourComponents, voxelHeight, nbhWalker);
			// decide W-Destructability for points with some lower neighbours
			if (lowerNeighbourComponents.empty())
				return NULL;
			SCompTreeNode<T> *highestFork = GetHighestFork(lowerNeighbourComponents);
			if (!highestFork)
				return *lowerNeighbourComponents.begin();
			if (highestFork->level < voxelHeight)
				return highestFork;
			return NULL;
		}

		/// constructs the labels image (minima labeled by non-zero number, non-minima by 0) for this component tree (result of the TW for this tree is in tWResult parameter)
		void ConstructLabelsImage(const Image3d<T> &tWResult, Image3d<size_t> *labels)
		{
			// go through all the points and mark minima with labels and non-minima with zeroes
			size_t imageSize = tWResult.GetImageSize();
			labels->CopyMetaData(tWResult);
			std::map<SCompTreeNode<T> *, size_t> labelMap;
			size_t nextLabel = 1;
			for (size_t i = 0; i < imageSize; ++i)
			{
				if (this->mapping[i]->children.first)
					labels->SetVoxel(i, 0);
				else
				{
					typename std::map<SCompTreeNode<T> *, size_t>::iterator it = labelMap.find(this->mapping[i]);
					if (it == labelMap.end())
					{
						labels->SetVoxel(i, nextLabel);
						labelMap[this->mapping[i]] = nextLabel++;
					}
					else
						labels->SetVoxel(i, it->second);
				}
			}
		}

	private:
		/// list of the nodes of the tree in depth first search
		SCompTreeNode<T> **dFSList;
		/// size of the dFSList
		size_t dFSListSize;
		/// table that holds precomputed information about the dFSList lowest element (according to the depth) index in some intervals (using the sparse table technique - m[i][j] holds the index of the lowest element for the interval of the length 2^j beginning at the i)
		size_t **minInd;

		/// creates the set of the components belonging to the neighbourhood pixels of the given one
		void GetLowerNeighbourComponents(const Image3d<T> &image, size_t index, std::set<SCompTreeNode<T>*, SCompTreeNodeCompareFunctor<T> > &lowerNeighbourComponents, T &voxelHeight, NeighbourhoodWalker &nbhWalker) const
		{
			// find the components of the lower neighbours
			voxelHeight = image.GetVoxel(index);
			for (nbhWalker.Begin(index); nbhWalker.Valid(); nbhWalker.Next())
			{
				size_t neighbour = nbhWalker.GetIndex();
				if (image.GetVoxel(neighbour) < voxelHeight)
					lowerNeighbourComponents.insert(this->mapping[neighbour]);
			}
		}
	};

	/// Priority queue class which preserves the order of the elements on the same level (don't know whether priority queue or multimap does this)
	template <typename T>
	class CTWQueue
	{
	public:
		/// returns whether is the queue empty
		bool Empty() const
		{
			return data.empty();
		}
		/// pushes the index into the queue with the specified priority
		void Push(size_t index, T priority)
		{
			data[priority].push_back(index);
		}
		/// removes the element with the lowest priority from the queue
		void Pop()
		{
			I3DASSERT_ALWAYS_ET(!data.empty(), "Trying to pop an alement from an empty queue.");
			typename std::map<T, std::deque<size_t> >::iterator i = data.begin();
			i->second.pop_front();
			if (i->second.empty())
				data.erase(i);
		}
		/// returns the element with the lowest priority
		size_t Top() const
		{
			I3DASSERT_ALWAYS_ET(!data.empty(), "Trying to retrieve an alement from an empty queue.");
			return data.begin()->second.front();
		}
		/// returns the priority of the element with the lowest priority
		T TopPriority() const
		{
			I3DASSERT_ALWAYS_ET(!data.empty(), "Trying to retrieve an alement from an empty queue.");
			return data.begin()->first;
		}

	private:
		/// stored values
		std::map<T, std::deque<size_t> > data;
	};

	/// Output operator that visualizes an instance of the SCompTreeTW
	template <typename T>
	I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTreeTW<T> &compTree)
	{
		return ost << &compTree;
	}

	/// Output operator that visualizes an instance of the SCompTreeTW
	template <typename T>
	I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTreeTW<T> *compTree)
	{
		std::ostream& result = ost << static_cast<const SCompTree<T>*>(compTree);
		compTree->PrintDFSList(ost);
		return result;
	}

	template <typename T>
	I3D_ALGO_EXPORT void WatershedTopological(const Image3d<T> &input, Image3d<T> &output, const Neighbourhood &neighbourhood, bool justMWS, Image3d<size_t> *labels)
	{
		I3DASSERT_ALWAYS_ET(input.GetImageSize() != 0, "Input image should not be empty.");

		// compute the "min-tree" (component tree of the inverted image) that is useful for computing the topological watershed
		Image3d<T> inputInversion;
		T maxInputValue = input.GetMaxValue();
		inputInversion.CopyMetaData(input);
		size_t imageSize = input.GetImageSize();
		for (size_t i = 0; i < imageSize; ++i)
			inputInversion.SetVoxel(i, maxInputValue - input.GetVoxel(i));
		SCompTreeTW<T> compTree(new CCompTreeNodeAttribHandlerTW<T>());
		NeighbourhoodWalker nbhWalker(input.GetSize(), neighbourhood, true);
		BuildComponentTree(inputInversion, compTree, nbhWalker);
		compTree.Finish(maxInputValue);
		// TODO we can probably simplify the tree - what about components with only one son? they can not be LCAs; updating of the component mapping is also needed
		// std::cout << compTree;
		// compute the M-watershed
		output = input;
		CTWQueue<T> queue;
		bool *inQueue = new bool[imageSize];
		memset(inQueue, 0, imageSize * sizeof(bool));
		for (size_t i = 0; i < imageSize; ++i)
		{
			if (compTree.MDestructible(output, i, nbhWalker))
			{
				queue.Push(i, output.GetVoxel(i));
				inQueue[i] = true;
			}
		}
		NeighbourhoodWalker::State nbhWalkerState;
		while (!queue.Empty())
		{
			size_t p = queue.Top();
			queue.Pop();
			inQueue[p] = false;
			SCompTreeNode<T> *targetComponent = compTree.MDestructible(output, p, nbhWalker);
			if (targetComponent)
			{
				output.SetVoxel(p, targetComponent->level);
				compTree.mapping[p] = targetComponent;
				for (nbhWalker.Begin(p); nbhWalker.Valid(); nbhWalker.Next())
				{
					size_t q = nbhWalker.GetIndex();
					nbhWalker.SaveState(nbhWalkerState);
					if (!inQueue[q] && compTree.MDestructible(output, q, nbhWalker))
					{
						queue.Push(q, output.GetVoxel(q));
						inQueue[q] = true;
					}
					nbhWalker.LoadState(nbhWalkerState);
				}
			}
		}
		delete[] inQueue;

		// quit if W-watershed is not required
		if (justMWS)
		{
			if (labels)
				compTree.ConstructLabelsImage(output, labels);
			return;
		}

		// finalize by computing the W-watershed
		SCompTreeNode<T> **newComponents = new SCompTreeNode<T>*[imageSize];
		memset(newComponents, 0, imageSize * sizeof(SCompTreeNode<T>*));
		T *newLevels = new T[imageSize];
		memset(newLevels, 0, imageSize * sizeof(T));
		SCompTreeNode<T> *targetComponent;
		for (size_t i = 0; i < imageSize; ++i)
		{
			if ((targetComponent = compTree.WDestructible(output, i, nbhWalker)))
			{
				queue.Push(i, targetComponent->level);
				newLevels[i] = targetComponent->level;
				newComponents[i] = targetComponent;
			}
		}
		while (!queue.Empty())
		{
			T k = queue.TopPriority();
			size_t p = queue.Top();
			queue.Pop();
			if (newLevels[p] == k)
			{
				output.SetVoxel(p, k);
				compTree.mapping[p] = newComponents[p];
				for (nbhWalker.Begin(p); nbhWalker.Valid(); nbhWalker.Next())
				{
					size_t q = nbhWalker.GetIndex();
					if (output.GetVoxel(q) <= k)
						continue;
					nbhWalker.SaveState(nbhWalkerState);
					targetComponent = compTree.WDestructible(output, q, nbhWalker);
					nbhWalker.LoadState(nbhWalkerState);
					if (targetComponent)
					{
						if (newLevels[q] != targetComponent->level)
						{
							queue.Push(q, targetComponent->level);
							newLevels[q] = targetComponent->level;
							I3DASSERT_BT(newLevels[q] >= k, "Why did I get a point that should have been processed before?");
							newComponents[q] = targetComponent;
						}
					}
					else
						newLevels[q] = std::numeric_limits<T>::max();
				}
			}
		}
		delete[] newLevels;
		delete[] newComponents;

		if (labels)
			compTree.ConstructLabelsImage(output, labels);
	}

	template I3D_ALGO_EXPORT void WatershedTopological(const Image3d<i3d::GRAY8> &input, Image3d<i3d::GRAY8> &output, const Neighbourhood &neighbourhood, bool justMWS, Image3d<size_t> *labels);
	template I3D_ALGO_EXPORT void WatershedTopological(const Image3d<i3d::GRAY16> &input, Image3d<i3d::GRAY16> &output, const Neighbourhood &neighbourhood, bool justMWS, Image3d<size_t> *labels);

} // namespace i3d
